#!/usr/bin/env python

"""
Tree modification functions
"""

import random


class TreeMod:
    """
    Return a tree with edge lengths modified according to one of 
    the mod functions. 
    """
    def __init__(self, toytree):
        self._toytree = toytree


    def node_scale_root_height(self, treeheight=1, include_stem=False, nocopy=False):
        """
        Returns a toytree copy with all nodes multiplied by a constant so that
        the root node height equals the value entered for treeheight. The 
        argument include_stem=True can be used to scale the tree so that the
        root + root.dist is equal to treeheight. This effectively sets the 
        stem height.
        """
        # make tree height = 1 * treeheight
        if nocopy:
            ctree = self._toytree
        else:
            ctree = self._toytree.copy()

        # get total tree height
        if include_stem:
            _height = ctree.treenode.height + ctree.treenode.dist
        else:
            _height = ctree.treenode.height

        # scale internal nodes 
        for node in ctree.treenode.traverse():
            node.dist = (node.dist / _height) * treeheight
        ctree._coords.update()
        return ctree


    def node_slider(self, prop=0.999, seed=None):
        """
        Returns a toytree copy with node heights modified while retaining 
        the same topology but not necessarily node branching order. 
        Node heights are moved up or down uniformly between their parent 
        and highest child node heights in 'levelorder' from root to tips.
        The total tree height is retained at 1.0, only relative edge
        lengths change.
        """
        # I don't think users should need to access prop
        prop = prop
        assert isinstance(prop, float), "prop must be a float"
        assert prop < 1, "prop must be a proportion >0 and < 1."
        random.seed(seed)

        # make copy and iter nodes from root to tips
        ctree = self._toytree.copy()
        for node in ctree.treenode.traverse():

            # slide internal nodes 
            if node.up and node.children:

                # get min and max slides
                # minjit = max([i.dist for i in node.children]) * prop
                # maxjit = (node.up.height * prop) - node.height

                # the closest child to me
                minchild = min([i.dist for i in node.children])

                # prop distance down toward child
                minjit = minchild * prop

                # prop towards parent
                maxjit = node.dist * prop

                # node.height
                newheight = random.uniform(
                    node.height - minjit, node.height + maxjit)

                # how much lower am i?
                delta = newheight - node.height

                # edges from children to reach me
                for child in node.children:
                    child.dist += delta

                # slide self to match
                node.dist -= delta

        # update new coords
        ctree._coords.update()
        return ctree


    def node_multiplier(self, multiplier=0.5, seed=None):
        """
        Returns a toytree copy with all nodes multiplied by a constant 
        sampled uniformly between (multiplier, 1/multiplier).
        """
        random.seed(seed)
        ctree = self._toytree.copy()
        low, high = sorted([multiplier, 1. / multiplier])
        mult = random.uniform(low, high)
        for node in ctree.treenode.traverse():
            node.dist = node.dist * mult
        ctree._coords.update()
        return ctree


    def make_ultrametric(self, strategy=1, nocopy=False):
        """
        Returns a tree with branch lengths transformed so that the tree is 
        ultrametric. Strategies include:

        (1) tip-align: 
            extend tips to the length of the fartest tip from the root; 
        (2) NPRS: 
            non-parametric rate-smoothing: minimize ancestor-descendant local 
            rates on branches to align tips (not yet supported); and 
        (3) penalized-likelihood: 
            not yet supported.
        """
        if nocopy:
            ctree = self._toytree
        else:
            ctree = self._toytree.copy()

        if strategy == 1:
            for node in ctree.treenode.traverse():
                if node.is_leaf():
                    node.dist += node.height
                    # node.dist = node.height + 1

        else:
            raise NotImplementedError(
                "Strategy {} not yet implemented. Seeking developers."
                .format(strategy))

        return ctree



#!/usr/bin/env python

"""
Robinson-Foulds Distance Calculations modified from ete3 TreeNode subfunction.
"""

from __future__ import print_function
from .utils import TreeError
# from .TreeParser import TreeParser


class RobinsonFoulds(object):
    """
    Simplified RF code for comparing TreeNodes.
    """
    def __init__(self, 
        t1,
        t2, 
        attr_t1="name", 
        attr_t2="name", 
        unrooted_trees=False, 
        expand_polytomies=False,
        polytomy_size_limit=5,
        skip_large_polytomies=False,
        correct_by_polytomy_size=False,
        min_support_t1=0.0,
        min_support_t2=0.0,
        ):

        # store attributes
        self.t1 = t1
        self.t2 = t2
        self.attr_t1 = attr_t1
        self.attr_t2 = attr_t2
        self.unrooted_trees = unrooted_trees
        self.expand_polytomies = expand_polytomies
        self.polytomy_size_limit = polytomy_size_limit
        self.skip_large_polytomies = skip_large_polytomies
        self.correct_by_polytomy_size = correct_by_polytomy_size
        self.min_support_t1 = min_support_t1
        self.min_support_t2 = min_support_t2

        # to be updated
        self.polytomy_correction = 0
        self.common_attrs = set()
        self.t1s = []
        self.t2s = []
        self.min_comparison = None

        # run functions
        self.check_args()
        self.check_attrs()
        self.get_trees()
        self.get_corrections()
        #self.compare_trees()


    def check_args(self):
        # check whether to bail out
        if self.unrooted_trees:
            if self.expand_polytomies:
                raise TreeError(
                    "Cannot use 'expand_polytomies' and 'unrooted_trees'")                    
        else:
            if (len(self.t1.children) > 2) or (len(self.t2.children) > 2):
                raise TreeError(
                    "Unrooted tree found! See 'unrooted_trees' flag.")

        if self.expand_polytomies and self.correct_by_polytomy_size:
            raise TreeError(
                "Cannot use expand_polytomies with correct_by_polytomy_size")


    def check_attrs(self):
        # get common attributes of the two trees, use generators
        attrs_t1 = (
            getattr(i, self.attr_t1) for i in self.t1.iter_leaves() 
            if hasattr(i, self.attr_t1)
        )
        attrs_t2 = (
            getattr(i, self.attr_t2) for i in self.t2.iter_leaves() 
            if hasattr(i, self.attr_t2)
        )
        self.common_attrs = set(attrs_t1) & set(attrs_t2)

        # Check for duplicated items (is this necessary?)
        size1 = sum((
            1 for i in self.t1.iter_leaves() if 
            getattr(i, self.attr_t1, None) in self.common_attrs
        ))
        size2 = sum((
            1 for i in self.t2.iter_leaves() if 
            getattr(i, self.attr_t2, None) in self.common_attrs
        ))
        if size1 > len(self.common_attrs):
            raise TreeError('Duplicated items found in source tree')
        if size2 > len(self.common_attrs):
            raise TreeError('Duplicated items found in reference tree')


    def get_trees(self):
        """
        TODO: rewrite this so it doesn't have to re-parse newicks.
        """
        # expand polytomies to get all resolutions possible, but fail if > max
        if self.expand_polytomies:
            raise NotImplementedError(
                "RF dist of unresolved trees not implemented currently. TODO: contact developers.")

            self.t1s = (
                TreeParser(nw).treenodes[0] for nw in
                self.t1.expand_polytomies(
                    map_attr=self.attr_t1,
                    polytomy_size_limit=self.polytomy_size_limit,
                    skip_large_polytomies=self.skip_large_polytomies
                    )
                )
            self.t2s = (
                # toytree.tree(nw).treenode for nw in                                 
                TreeParser(nw).treenodes[0] for nw in 
                self.t2.expand_polytomies(
                    map_attr=self.attr_t2,
                    polytomy_size_limit=self.polytomy_size_limit,
                    skip_large_polytomies=self.skip_large_polytomies,
                    )
                )
            # why ...?
            self.attr_t1, self.attr_t2 = "name", "name"
        else:
            self.t1s = [self.t1]
            self.t2s = [self.t2]


    def get_corrections(self):
        # correct for the N polytomies
        if self.correct_by_polytomy_size:
            corr1 = (
                len(i.children) - 2 for i in self.t1.traverse() 
                if len(i.children) > 2
            )
            corr2 = (
                len(i.children) - 2 for i in self.t2.traverse() 
                if len(i.children) > 2
            )
            if corr1 and corr2:
                raise TreeError(
                    "Both trees have polytomies! Try expand_polytomies=True")
            else:
                self.polytomy_correction = max((corr1, corr2))


    def get_edges(self, tx_content, tx_leaves, attr):

        if self.unrooted_trees:
            edges = set()
            for content in tx_content.values():  
                    
                # get name from nodes if they have a name and its in common
                names1 = tuple(sorted(
                    getattr(node, attr) for node in content if 
                    (
                        hasattr(node, attr) and \
                        (getattr(node, attr) in self.common_attrs)
                    )
                ))

                # get names of all leaves minus this node's descendants
                names2 = tuple(sorted(
                    getattr(node, attr) for node in tx_leaves - content if 
                    (
                        hasattr(node, attr) and \
                        (getattr(node, attr) in self.common_attrs)
                    )
                ))

                # add the split to the edges set
                edges.add(tuple(sorted(set([names1, names2]))))
            edges.discard(((), ()))

        # get edges of a rooted tree: just tuples of tips descended
        else:
            edges = set()
            for content in tx_content.values():
                names1 = tuple(sorted(
                    getattr(node, attr) for node in content if 
                    (
                        hasattr(node, attr) and \
                        (getattr(node, attr) in self.common_attrs)
                    )
                ))
                edges.add(tuple(sorted(set(names1))))
            edges.discard(())
        return edges


    def get_support_dict(self, tx_content, attr):
        cdict = {}
        for branch, content in tx_content.items():
            key = tuple(sorted(
                getattr(node, attr) for node in content if 
                (
                    hasattr(node, attr) and \
                    (getattr(node, attr) in self.common_attrs)
                )
            ))
            cdict[key] = branch.support
        return cdict


    def get_discards(self, t1_edges, t1_sdict, t2_edges, t2_sdict):

        # initial empty
        discard_t1, discard_t2 = set(), set()

        # get discards from t1
        if self.min_support_t1 and self.unrooted_trees:
            discard_t1 = set()
            for split in t1_edges:
                split_support = t1_sdict.get(
                    split[0], 
                    t1_sdict.get(split[1], 999999999)
                )
                if split_support < self.min_support_t1:
                    discard_t1.add(split)

        elif self.min_support_t1:
            discard_t1 = set([
                split for split in t1_edges 
                if t1_sdict[split] < self.min_support_t1
            ])

        # get discards from t2
        if self.min_support_t2 and self.unrooted_trees:
            discard_t2 = set()
            for split in t2_edges:
                split_support = t2_sdict.get(
                    split[0], 
                    t2_sdict.get(split[1], 999999999)
                )
                if split_support < self.min_support_t2:
                    discard_t2.add(split)

        elif self.min_support_t2:
            discard_t2 = set([
                split for split in t2_edges 
                if t2_sdict[split] < self.min_support_t2
            ])
        return discard_t1, discard_t2


    def compare_trees(self):
        """
        Iterate over trees in t1 and t2 to count splits present in both
        """
        for t1 in self.t1s:
            # a dictionary of {node: node.get_leaves()}
            t1_content = t1.get_cached_content()

            # a list of nodes descended from this one: [node0, node1, node2]
            t1_leaves = t1_content[t1]
            
            # get edges of the tree: set of tuples on either side of splits
            t1_edges = self.get_edges(t1_content, t1_leaves, self.attr_t1)

            # get support on tree ...
            t1_sdict = None
            if self.min_support_t1:
                t1_sdict = self.get_support_dict(t1_content, self.attr_t1)

            # iterate over target trees
            for t2 in self.t2s:
                # a dictionary of {node: node.get_leaves()}
                t2_content = t2.get_cached_content()
                
                # nodes descended from this one: [node0, node1, node2, ...]
                t2_leaves = t2_content[t2]

                # get edges of the tree: set of tuples on either side of splits
                t2_edges = self.get_edges(t2_content, t2_leaves, self.attr_t2)

                # get support dict
                t2_sdict = None
                if self.min_support_t2:
                    t2_sdict = self.get_support_dict(t2_content, self.attr_t2)

                # if support constraint, discard lowly supported splits
                discard_t1, discard_t2 = self.get_discards(
                    t1_edges, t1_sdict, t2_edges, t2_sdict)

                # the two root edges are never counted here, as they are always
                # present in both trees because of the common attr filters
                rf = len(((t1_edges ^ t2_edges) - discard_t2) - discard_t1)
                rf -= self.polytomy_correction

                # count parts
                cedges1 = t1_edges - discard_t1
                cedges2 = t2_edges - discard_t2

                if self.unrooted_trees:
                    max_parts = sum((
                        sum(1 for split in cedges1 if split[0] and split[1]),
                        sum(1 for split in cedges2 if split[0] and split[1]),
                    ))
                else:
                    # Otherwise we need to count the actual number of valid
                    # partitions in each tree -2 is to avoid counting the root
                    # partition of the two trees (only needed in rooted trees)
                    max_parts = sum((
                        sum(1 for split in cedges1 if split),
                        sum(1 for split in cedges2 if split),
                    )) - 2

                # update min_comparison if this compare was worse
                if not self.min_comparison or (self.min_comparison[0] > rf):
                    min_comparison = [
                        rf, 
                        max_parts, 
                        self.common_attrs, 
                        t1_edges, 
                        t2_edges, 
                        discard_t1, 
                        discard_t2,
                    ]

        return min_comparison


#!/usr/bin/env python

"""
Random Tree generation Class
"""

import random
import toytree
from .utils import ToytreeError


class RandomTree(object):

    @staticmethod
    def coaltree(ntips, ne=None, seed=None):
        """
        Returns a coalescent tree with ntips samples and waiting times 
        between coalescent events drawn from the kingman coalescent:
        (4N)/(k*(k-1)), where N is population size and k is sample size.
        Edge lengths on the tree are in generations.

        If no Ne argument is entered then edge lengths are returned in units
        of 2*Ne, i.e., coalescent time units. 
        """
        # seed generator
        random.seed(seed)

        # convert units
        coalunits = False
        if not ne:
            coalunits = True
            ne = 10000

        # build tree: generate N tips as separate Nodes then attach together 
        # at internal nodes drawn randomly from coalescent waiting times.
        tips = [
            toytree.tree().treenode.add_child(name=str(i)) 
            for i in range(ntips)
        ]
        while len(tips) > 1:
            rtree = toytree.tree()
            tip1 = tips.pop(random.choice(range(len(tips))))
            tip2 = tips.pop(random.choice(range(len(tips))))
            kingman = (4. * ne) / float(ntips * (ntips - 1))
            dist = random.expovariate(1. / kingman)
            rtree.treenode.add_child(tip1, dist=tip2.height + dist)
            rtree.treenode.add_child(tip2, dist=tip1.height + dist)
            tips.append(rtree.treenode)

        # build new tree from the newick string
        self = toytree.tree(tips[0].write())    
        self.treenode.ladderize()

        # make tree edges in units of 2N (then N doesn't matter!)
        if coalunits:
            for node in self.treenode.traverse():
                node.dist /= (2. * ne)

        # ensure tips are at zero (they sometime vary just slightly)
        for node in self.treenode.traverse():
            if node.is_leaf():
                node.dist += node.height

        # set tipnames
        for tip in self.get_tip_labels():
            node = self.treenode.search_nodes(name=tip)[0]
            node.name = "r{}".format(node.idx)

        # decompose fills in internal node names and idx
        self._coords.update()
        return self


    @staticmethod
    def unittree(ntips, treeheight=1.0, seed=None):
        """
        Returns a random tree ultrametric topology w/ N tips and a root 
        height set to 1 or a user-entered treeheight value. Descendant 
        nodes are evenly spaced between the root and time 0.

        Parameters
        -----------
        ntips (int):
            The number of tips in the randomly generated tree

        treeheight(float):
            Scale tree height (all edges) so that root is at this height.

        seed (int):
            Random number generator seed.
        """
        # seed generator
        random.seed(seed)

        # generate tree with N tips.
        tmptree = toytree.tree().treenode  # TreeNode()
        tmptree.populate(ntips)
        self = toytree.tree(newick=tmptree.write())

        # set tip names by labeling sequentially from 0
        self = (
            self
            .ladderize()
            .mod.make_ultrametric()
            .mod.node_scale_root_height(treeheight)
        )

        # set tipnames randomly (doesn't have to match idx)
        nidx = list(range(self.ntips))
        random.shuffle(nidx)
        for tidx, node in enumerate(self.treenode.get_leaves()):
            node.name = "r{}".format(nidx[tidx])

        for node in self.treenode.traverse():
            node.support = 100            
        # fill internal node names and idx
        self.treenode.ladderize()
        self._coords.update()
        return self


    @staticmethod
    def imbtree(ntips, treeheight=1.0):
        """
        Return an imbalanced (comb-like) tree topology.
        """
        node = toytree.TreeNode.TreeNode()
        node.add_child(name="r0")
        node.add_child(name="r1")

        for i in range(2, ntips):
            # empty node
            cherry = toytree.TreeNode.TreeNode()
            # add new child
            cherry.add_child(name="r" + str(i))
            # add old tree
            cherry.add_child(node)
            # update rtree
            node = cherry

        # get toytree from newick            
        tre = toytree.tree(node)
        tre = tre.mod.make_ultrametric(nocopy=True)
        tre = tre.mod.node_scale_root_height(treeheight, nocopy=True)
        tre._coords.update()
        return tre


    @staticmethod
    def baltree(ntips, treeheight=1.0):
        """
        Returns a balanced tree topology.
        """
        # require even number of tips
        if ntips % 2:
            raise ToytreeError("balanced trees must have even number of tips.")

        # make first cherry
        rtree = toytree.tree()
        rtree.treenode.add_child(name="r0")
        rtree.treenode.add_child(name="r1")

        # add tips in a balanced way
        for i in range(2, ntips):

            # get node to split
            node = return_small_clade(rtree.treenode)

            # add two children
            node.add_child(name=node.name)
            node.add_child(name="r" + str(i))

            # rename ancestral node
            node.name = None

        # rename tips so names are in order
        idx = len(rtree) - 1
        for node in rtree.treenode.traverse("postorder"):
            if node.is_leaf():
                node.name = "r" + str(idx)
                idx -= 1

        # get toytree from newick            
        tre = toytree.tree(rtree.write(tree_format=9))
        tre = tre.mod.make_ultrametric()
        self = tre.mod.node_scale_root_height(treeheight)
        self._coords.update()
        return self        


    @staticmethod
    def rtree(ntips, treeheight=1.0, seed=None):
        """
        Randomly assigns edge lengths between U(0-1) to edges 
        and then scales total treeheight to 1.0 or entered value.
        """
        # get a unit tree
        self = toytree.rtree.unittree(ntips, treeheight, seed)      

        # randomly assign node dists
        self = self.set_node_values(
            "dist", {i: random.random() for i in range(self.nnodes)}
        )

        # rescale total height to .
        self = self.mod.node_scale_root_height(treeheight, nocopy=True)
        return self        


def return_small_clade(treenode):
    "used to produce balanced trees, returns a tip node from the smaller clade"
    node = treenode
    while 1:
        if node.children:
            c1, c2 = node.children
            node = sorted([c1, c2], key=lambda x: len(x.get_leaves()))[0]
        else:
            return node



#!/usr/bin/env python

"""
A newick/nexus file/string parser based on the ete3.parser.newick. Takes as 
input a string or file that contains one or multiple lines that contain 
newick strings. Lines that do not contain newick strings are ignored, unless
the #NEXUS is in the header in which case the 'translate' and 'trees' blocks
are parsed. 
"""

import os
import re
import requests
from .TreeNode import TreeNode
from .utils import NW_FORMAT

# Regular expressions used for reading newick format
FLOAT_RE = r"\s*[+-]?\d+\.?\d*(?:[eE][-+]\d+)?\s*"
NAME_RE = r"[^():,;]+?"
NHX_RE = r"\[&&NHX:[^\]]*\]"
MB_BRLEN_RE = r"\[&B (\w+) [0-9.e-]+\]"


class NewickError(Exception):
    """Exception class designed for NewickIO errors."""
    def __init__(self, value):
        Exception.__init__(self, value)


class NexusError(Exception):
    """Exception class designed for NewickIO errors."""
    def __init__(self, value):
        Exception.__init__(self, value)



class FastTreeParser():
    """
    A less flexible but faster newick parser for performance sensitive apps.
    Only supports newick string input in format 0.
    """
    def __init__(self, newick, tree_format):
        self.data = newick
        extractor = FastNewick2TreeNode(self.data, tree_format)
        self.treenode = extractor.newick_from_string()



class TreeParser(object):
    def __init__(self, intree, tree_format=0, multitree=False, debug=False):
        """
        Reads input as a string or file, figures out format and parses it.
        Formats 0-10 are newick formats supported by ete3. 
        Format 11 is nexus format from mrbayes.

        Returns either a Toytree or MultiTree object, depending if input has
        one or more trees. 
        """
        # the input file/stream and the loaded data
        self.intree = intree
        self.data = None
        self.debug = debug

        # the tree_format and parsed tree string from data
        self.fmt = tree_format
        self.multitree = multitree
        self.newick = ""

        # returned result: 1 tree for Toytree multiple trees for MultiTrees
        self.treenodes = []

        # newick translation dictionary
        self.tdict = {}

        # compiled re matchers for this tree format type
        self.matcher = MATCHER[self.fmt]

        # parse intree
        if not self.debug:
            self._run()


    def _run(self):
        # get newick from data and test newick structure
        if self.intree:
            # read data input by lines to .data
            self.get_data_from_intree()

            # check for NEXUS wrappings and update .data for newick strings
            self.parse_nexus()

            # raise warnings if tree_format doesn't seem right for data
            self.warn_about_format()

            # parse newick strings to treenodes list
            self.get_treenodes()

            # apply names from tdict
            self.apply_name_translation()

        # no input data
        else:
            self.treenodes = [TreeNode()]


    def warn_about_format(self):
        # warning about formats
        if "[&&NHX" not in self.data[0]:
            if ("[&" in self.data[0]) & (self.fmt != 10):
                print("Warning: data looks like tree_format=10 (mrbayes-like)")          


    def get_data_from_intree(self):
        """
        Load *data* from a file or string and return as a list of strings.
        The data contents could be one newick string; a multiline NEXUS format
        for one tree; multiple newick strings on multiple lines; or multiple
        newick strings in a multiline NEXUS format. In any case, we will read
        in the data as a list on lines. 
        """

        # load string: filename or data stream
        if isinstance(self.intree, (str, bytes)):

            # strip it
            self.intree = self.intree.strip()

            # is a URL: make a list by splitting a string
            if any([i in self.intree for i in ("http://", "https://")]):
                response = requests.get(self.intree)
                response.raise_for_status()
                self.data = response.text.strip().split("\n")

            # is a file: read by lines to a list
            elif os.path.exists(self.intree):
                with open(self.intree, 'rU') as indata:
                    self.data = indata.readlines()

            # is a string: make into a list by splitting
            else:
                self.data = self.intree.split("\n")

        # load iterable: iterable of newick strings
        elif isinstance(self.intree, (list, set, tuple)):
            self.data = list(self.intree)


    def parse_nexus(self):
        "get newick data from NEXUS"
        if self.data[0].strip().upper() == "#NEXUS":
            nex = NexusParser(self.data)
            self.data = nex.newicks
            self.tdict = nex.tdict


    def get_treenodes(self):
        "test format of intree nex/nwk, extra features"

        if not self.multitree:
            # get TreeNodes from Newick
            extractor = Newick2TreeNode(self.data[0].strip(), fmt=self.fmt)

            # extract one tree
            self.treenodes.append(extractor.newick_from_string())

        else:
            for tre in self.data:
                # get TreeNodes from Newick
                extractor = Newick2TreeNode(tre.strip(), fmt=self.fmt)

                # extract one tree
                self.treenodes.append(extractor.newick_from_string())


    def apply_name_translation(self):
        if self.tdict:
            for tree in self.treenodes:
                for node in tree.traverse():
                    if node.name in self.tdict:
                        node.name = self.tdict[node.name]



class Newick2TreeNode:
    "Parse newick str to a TreeNode object"    
    def __init__(self, data, fmt=0):
        self.data = data
        self.root = TreeNode()
        self.current_node = self.root
        self.current_parent = None
        self.fmt = fmt
        self.cleanup_data()


    def cleanup_data(self):
        # check parentheses
        if self.data.count('(') != self.data.count(')'):
            raise NewickError('Parentheses do not match. Broken tree data.')

        # remove white spaces and separators 
        self.data = re.sub(r"[\n\r\t ]+", "", self.data)

        # mrbayes format terrible formatting hacks--------
        if self.fmt == 10:

            # convert bracket markers to NHX format
            self.data = self.data.replace("[&", "[&&NHX:")

            # replace commas inside feature strings with dashes
            ns = ""
            for chunk in self.data.split("{"):
                if "}" in chunk:
                    pre, post = chunk.split("}", 1)
                    pre = pre.replace(",", "-")
                    ns += "{" + pre + "}" + post
                else:
                    ns += chunk
            self.data = ns

            # replace parentheses inside brackets with curly braces
            ns = ""
            for chunk in self.data.split("["):
                if "]" in chunk:
                    pre, post = chunk.split("]", 1)
                    pre = pre.replace("(", "{")
                    pre = pre.replace(")", "}")
                    pre = pre.replace(",", ":")
                    pre = pre.replace('"', "")
                    pre = pre.replace("'", "")
                    ns += "[" + pre + "]" + post
                else:
                    ns += chunk
            self.data = ns            


    def newick_from_string(self):
        "Reads a newick string in the New Hampshire format."

        # split on parentheses to traverse hierarchical tree structure
        for chunk in self.data.split("(")[1:]:
            # add child to make this node a parent.
            self.current_parent = (
                self.root if self.current_parent is None else
                self.current_parent.add_child()
            )

            # get all parenth endings from this parenth start
            subchunks = [ch.strip() for ch in chunk.split(",")]
            if subchunks[-1] != '' and not subchunks[-1].endswith(';'):
                raise NewickError(
                    'Broken newick structure at: {}'.format(chunk))

            # Every closing parenthesis will close a node and go up one level.
            for idx, leaf in enumerate(subchunks):
                if leaf.strip() == '' and idx == len(subchunks) - 1:
                    continue
                closing_nodes = leaf.split(")")

                # parse features and apply to the node object
                self.apply_node_data(closing_nodes[0], "leaf")

                # next contain closing nodes and data about the internal nodes.
                if len(closing_nodes) > 1:
                    for closing_internal in closing_nodes[1:]:
                        closing_internal = closing_internal.rstrip(";")
                        # read internal node data and go up one level
                        self.apply_node_data(closing_internal, "internal")
                        self.current_parent = self.current_parent.up
        return self.root


    def apply_node_data(self, subnw, node_type):

        if node_type in ("leaf", "single"):
            self.current_node = self.current_parent.add_child()
        else:
            self.current_node = self.current_parent

        # if no feature data
        subnw = subnw.strip()
        if not subnw:
            return 

        # load matcher junk
        c1, c2, cv1, cv2, match = MATCHER[self.fmt].type[node_type]

        # if beast or mb then combine brackets
        if self.fmt == 10:
            if "]:" not in subnw:
                node, edge = subnw.split("]", 1)
                subnw = node + "]:0.0" + edge                       
            node, edge = subnw.split("]:")
            npre, npost = node.split("[")

            # mrbayes mode: (a[&a:1,b:2]:0.1[&c:10])
            try:
                epre, epost = edge.split("[")
                subnw = "{}:{}[&&NHX:{}".format(
                    npre, epre, ":".join([npost[6:], epost[6:]]))

            # BEAST mode: (a[&a:1,b:2,c:10]:0.1)
            except ValueError:
                subnw = "{}:{}[&&NHX:{}]".format(npre, edge, npost[6:])

        # look for node features
        data = re.match(match, subnw)

        # if there are node features then add them to this node
        if data:
            data = data.groups()

            # data should not be empty
            if all([i is None for i in data]):
                raise NewickError(
                    "Unexpected newick format {}".format(subnw))

            # node has a name
            if (data[0] is not None) and (data[0] != ''):
                self.current_node.add_feature(c1, cv1(data[0]))

            if (data[1] is not None) and (data[1] != ''):
                self.current_node.add_feature(c2, cv2(data[1][1:]))

            if (data[2] is not None) and data[2].startswith("[&&NHX"):
                fdict = parse_nhx(data[2])
                for fname, fvalue in fdict.items():
                    self.current_node.add_feature(fname, fvalue)
        else:
            raise NewickError("Unexpected newick format {}".format(subnw))



class NexusParser:
    """
    Parse nexus file/str formatted data to extract tree data and features.
    Expects '#NEXUS', 'begin trees', 'tree', and 'end;'.
    """
    def __init__(self, data, debug=False):

        self.data = data
        self.newicks = []
        self.tdict = {}
        self.matcher = re.compile(MB_BRLEN_RE)
        if not debug:
            self.extract_tree_block()


    def extract_tree_block(self):
        "iterate through data file to extract trees"        

        # data SHOULD be a list of strings at this point
        lines = iter(self.data)
        while 1:
            try:
                line = next(lines).strip()
            except StopIteration:
                break

            # oh mrbayes, you seriously allow spaces within newick format!?
            # find "[&B TK02Brlens 8.123e-3]" and change to [&Brlen=8.123e-3]
            # this is a tmp hack fix, to be replaced with a regex
            line = line.replace(" TK02Brlens ", "=")

            # enter trees block
            if line.lower() == "begin trees;":
                while 1:
                    # iter through trees block
                    nextline = next(lines).strip()

                    # remove horrible brlen string with spaces from mb
                    nextline = self.matcher.sub("", nextline)

                    # split into parts on spaces
                    sub = nextline.split()

                    # skip if a blank line
                    if not sub:
                        continue

                    # look for translation
                    elif sub[0].lower() == "translate":
                        while not sub[-1].endswith(";"):
                            sub = next(lines).strip().split()
                            self.tdict[sub[0]] = sub[-1].strip(",").strip(";")

                    # parse tree blocks
                    elif sub[0].lower().startswith("tree"):
                        self.newicks.append(sub[-1])

                    # end of trees block
                    elif sub[0].lower() == "end;":
                        break



# re matchers should all be compiled on toytree init
class Matchers:
    def __init__(self, formatcode):
        self.type = {}

        for node_type in ["leaf", "single", "internal"]:
            # (node_type == "leaf") or (node_type == "single"):
            if node_type != "internal":  
                container1 = NW_FORMAT[formatcode][0][0]
                container2 = NW_FORMAT[formatcode][1][0]
                converterFn1 = NW_FORMAT[formatcode][0][1]
                converterFn2 = NW_FORMAT[formatcode][1][1]
                flexible1 = NW_FORMAT[formatcode][0][2]
                flexible2 = NW_FORMAT[formatcode][1][2]
            else:
                container1 = NW_FORMAT[formatcode][2][0]
                container2 = NW_FORMAT[formatcode][3][0]
                converterFn1 = NW_FORMAT[formatcode][2][1]
                converterFn2 = NW_FORMAT[formatcode][3][1]
                flexible1 = NW_FORMAT[formatcode][2][2]
                flexible2 = NW_FORMAT[formatcode][3][2]

            if converterFn1 == str:
                first_match = "({})".format(NAME_RE)
            elif converterFn1 == float:
                first_match = "({})".format(FLOAT_RE)
            elif converterFn1 is None:
                first_match = '()'

            if converterFn2 == str:
                second_match = "(:{})".format(NAME_RE)
            elif converterFn2 == float:
                second_match = "(:{})".format(FLOAT_RE)
            elif converterFn2 is None:
                second_match = '()'

            if flexible1 and (node_type != 'leaf'):
                first_match += "?"
            if flexible2:
                second_match += "?"

            m0 = r"^\s*{first_match}"
            m1 = r"\s*{second_match}"
            m2 = r"\s*({NHX_RE})"
            m3 = r"?\s*$"
            matcher_str = (m0 + m1 + m2 + m3).format(**{
                "first_match": first_match, 
                "second_match": second_match,
                "NHX_RE": NHX_RE,
                })

            # matcher_str = r'^\s*{}\s*{}\s*({})?\s*$'.format(
            # FIRST_MATCH, SECOND_MATCH, NHX_RE)
            compiled_matcher = re.compile(matcher_str)

            # fill matcher for this node
            self.type[node_type] = [
                container1,
                container2, 
                converterFn1, 
                converterFn2, 
                compiled_matcher
            ]



def parse_messy_nexus(nexus):
    """
    Approaches: 
    1: Parse the format to match NHX and then parse like normal NHX. 
    2. New parser to store all features and values as strings. 
    """


def parse_nhx(NHX_string):
    """ 
    NHX format:  [&&NHX:prop1=value1:prop2=value2] 
    MB format: ((a[&Z=1,Y=2], b[&Z=1,Y=2]):1.0[&L=1,W=0], ...
    """
    # store features
    ndict = {}

    # parse NHX or MB features
    if "[&&NHX:" in NHX_string:
        NHX_string = NHX_string.replace("[&&NHX:", "")
        NHX_string = NHX_string.replace("]", "")

        for field in NHX_string.split(":"):
            try:
                pname, pvalue = field.split("=")
                ndict[pname] = pvalue
            except ValueError as e:
                raise NewickError('Invalid NHX format %s' % field)
    return ndict


# GLOBAL RE COMPILED MATCHERS
MATCHER = {}
for formatcode in range(11):
    MATCHER[formatcode] = Matchers(formatcode)



class FastNewick2TreeNode:
    "Parse newick str to a TreeNode object"    
    def __init__(self, data, tree_format):
        self.data = data
        self.root = TreeNode()
        self.current_node = self.root
        self.current_parent = None
        self.fmt = tree_format
        self.data = re.sub(r"[\n\r\t ]+", "", self.data)



    def newick_from_string(self):
        "Reads a newick string in the New Hampshire format."

        # split on parentheses to traverse hierarchical tree structure
        for chunk in self.data.split("(")[1:]:
            # add child to make this node a parent.
            self.current_parent = (
                self.root if self.current_parent is None else
                self.current_parent.add_child()
            )

            # get all parenth endings from this parenth start
            subchunks = [ch.strip() for ch in chunk.split(",")]
            if subchunks[-1] != '' and not subchunks[-1].endswith(';'):
                raise NewickError(
                    'Broken newick structure at: {}'.format(chunk))

            # Every closing parenthesis will close a node and go up one level.
            for idx, leaf in enumerate(subchunks):
                if leaf.strip() == '' and idx == len(subchunks) - 1:
                    continue
                closing_nodes = leaf.split(")")

                # parse features and apply to the node object
                self.apply_node_data(closing_nodes[0], "leaf")

                # next contain closing nodes and data about the internal nodes.
                if len(closing_nodes) > 1:
                    for closing_internal in closing_nodes[1:]:
                        closing_internal = closing_internal.rstrip(";")
                        # read internal node data and go up one level
                        self.apply_node_data(closing_internal, "internal")
                        self.current_parent = self.current_parent.up
        return self.root



    def apply_node_data(self, subnw, node_type):

        if node_type in ("leaf", "single"):
            self.current_node = self.current_parent.add_child()
        else:
            self.current_node = self.current_parent

        # if no feature data
        subnw = subnw.strip()
        if not subnw:
            return 

        # load matcher junk
        c1, c2, cv1, cv2, match = MATCHER[self.fmt].type[node_type]

        # look for node features
        data = re.match(match, subnw)

        # if there are node features then add them to this node
        if data:
            data = data.groups()

            # node has a name
            if (data[0] is not None) and (data[0] != ''):
                self.current_node.add_feature(c1, cv1(data[0]))

            if (data[1] is not None) and (data[1] != ''):
                self.current_node.add_feature(c2, cv2(data[1][1:]))

        else:
            raise NewickError("Unexpected newick format {}".format(subnw))



#!/usr/bin/env python

"""
Rooting class
"""

from .NodeAssist import NodeAssist
from .utils import ToytreeError


class Rooter:
    def __init__(self, tree, nastuple, resolve_root_dist, edge_features):
        """
        See docstring in ToyTree.root()
        """
        # store args
        self.tree = tree
        self.resolve_root_dist = resolve_root_dist
        self.edge_features = edge_features
        self.features = {"name", "dist", "support"}
        self.nas = None

        # dict for relabeling nodes {node: [parent, child, feat]}
        self.tdict = {}

        # the new node that will be inserted
        self.nnode = None

        # make a copy and ensure supports are either all int or float
        self.maxsup = max([int(i.support) for i in self.tree.treenode.traverse()])
        self.maxsup = (1.0 if self.maxsup <= 1.0 else 100)
        self.get_features()

        # parse node selecting arguments (nastuple) with NodeAssist
        self.get_match(*nastuple)

        # get the mrca node (or tip node) of the monopyletic matched query.
        self.node1 = self.nas.get_mrca()

        # the node on the other side of the edge to be split.
        self.node2 = self.node1.up

        # if rooting where root already exists then return current tree
        x0 = (self.node1.is_root())
        x1 = (self.node2.is_root() and self.tree.is_rooted())
        if not (x0 or x1):           

            # create new root node on an existing edge to split it.
            self.insert_new_node()

            # update edge lengths given new node insertion
            self.config_root_dist()

            # redirect support values and possibly others
            self.redirect_edge_features()
            self.restructure_tree()

            # update coodrds on tree
            self.update_tree_from_tdict()
            self.update()



    def update_tree_from_tdict(self):
        # update tree structure and node labels
        for node in self.tdict:
            node.up = self.tdict[node][0]
            node.children = self.tdict[node][1]
            for key, val in self.tdict[node][2].items():
                setattr(node, key, val)


    def update(self):
        # update coordinates which updates idx and adds it to any new nodes.
        self.tree.treenode = self.nnode
        self.tree.treenode.ladderize()
        self.tree._coords.update()



    def redirect_edge_features(self):
        """
        Set support values to maximum for new node since the user forced
        rooting, i.e, it is not uncertain. 
        """
        # mark new split with zero...
        for feature in set(self.edge_features) - set(["support", "dist"]):
            self.tdict[self.node2][2][feature] = 0.0

        # unless support value, then mark with full.
        if "support" in self.edge_features:
            self.tdict[self.node2][2]['support'] = self.maxsup
        else:
            self.tdict[self.node2][2]['support'] = self.node2.support



    def restructure_tree(self):
        """
        At this point tdict 
           (node): (parent) (children), features
        {
            nnode: [None, [node1, node2], {}]
            node1: [nnode, node1.children, {'dist'}]
            node2: [nnode, node2.children, {'dist': 0.0}]
        }
        """
        # start with the node leading from new root child2 to the 
        # rest of the tree structure and move up until old root.
        tnode = self.node2.up

        # label all remaining nodes by moving up from tnode to old root.
        while 1:

            # early break
            if not tnode:
                break

            # get parent node (should be already in tdict)
            # and children to be mod'd (the ones not yet in tdict)
            parent = [i for i in tnode.children if i in self.tdict][0]
            children = [i for i in tnode.children if i not in self.tdict]

            # break occurs after tnode is root
            if tnode.is_root():

                # need a root add feature here if unrooted...
                if len(children) > 1:

                    # update dist from new parent
                    self.tdict[tnode] = [
                        parent, 
                        children, 
                        {"dist": parent.dist},
                    ]

                    # update edge features from new parent
                    for feature in self.edge_features:
                        self.tdict[tnode][2][feature] = getattr(parent, feature)

                    # set tnode as parent's new child
                    self.tdict[parent][1].append(tnode)

                    # set children as descendant from tnode
                    for child in children:
                        self.tdict[child] = [tnode, child.children, {}]

                # get children that are not in tdict yet
                else:
                    for child in children:

                        # record whose children they are now 
                        # (node2 already did this)
                        if parent is self.node2:
                            self.tdict[self.node2][1].append(child)
                        else:
                            self.tdict[parent][1].append(child)

                        # record whose parents they have now and find distance
                        dist = {"dist": sum([i.dist for i in tnode.children])}
                        self.tdict[child] = [parent, child.children, dist]

                # finished
                break

            # normal nodes
            else:
                # update tnode.features [dist will be inherited from child]
                features = {'dist': tnode.dist, 'support': tnode.support}

                # keep connecting swap parent-child up to root
                if not tnode.up.is_root():
                    children += [tnode.up]

                # pass support values down (up in new tree struct)
                child = [i for i in tnode.children if i in self.tdict][0]                
                for feature in {'dist'}.union(self.edge_features):                   
                    features[feature] = getattr(child, feature)

                # store node update vals
                self.tdict[tnode] = [parent, children, features]

            # move towards root
            tnode = tnode.up



    def config_root_dist(self):
        """
        Now that the new root node is inserted .dist features must be 
        set for the two descendant nodes. Midpoint rooting is a common 
        option, but users can toggle 'resolve_root_dist' to change this.
        """
        # if not already at root polytomy, then connect node2 to parent
        if self.node2.up:
            if not self.node2.up.is_root():
                self.tdict[self.node2][1] += [self.node2.up]

        # if False create zero length root node
        if self.resolve_root_dist is False:
            self.resolve_root_dist = 0.0

        # if True then use midpoint rooting
        elif self.resolve_root_dist is True:
            self.tdict[self.node1][2]["dist"] = self.node1.dist / 2.
            self.tdict[self.node2][2]["dist"] = self.node1.dist / 2.

        # split the edge on 0 or a float
        if isinstance(self.resolve_root_dist, float):
            self.tdict[self.node1][2]["dist"] = self.node1.dist - self.resolve_root_dist            
            self.tdict[self.node2][2]["dist"] = self.resolve_root_dist
            if self.resolve_root_dist > self.node1.dist:
                raise ToytreeError("\n"
                "To preserve existing edge lengths the 'resolve_root_dist' arg\n"
                "must be smaller than the edge being split (it is selecting a \n"
                "a point along the edge.) The edge above node idx {} is {}."
                .format(self.node1.idx, self.node1.dist)
                )



    def insert_new_node(self):
        """

        """
        # the new root node to be placed on the split
        self.nnode = self.tree.treenode.__class__()
        self.nnode.name = "root"
        self.nnode.add_feature("idx", self.tree.treenode.idx)
        self.nnode.support = self.maxsup

        # remove node1 lineage leaving just node2 branch to be made into child
        self.node2.children.remove(self.node1)

        # new node has no parent and 1,2 as children and default features
        self.tdict[self.nnode] = [None, [self.node1, self.node2], {}]

        # node1 has new root parent, same children, and dist preserved 
        # (or split?), or should: node1.dist / 2.
        self.tdict[self.node1] = [
            self.nnode, 
            self.node1.children, 
            {"dist": self.node1.dist}
        ]  

        # node2 has new root parent, same children + mods, and dist/supp mods
        self.tdict[self.node2] = [
            self.nnode,
            self.node2.children,
            {"dist": 0.0},
        ]



    def get_features(self):
        # define which features to use/keep on nodes and which are "edge" 
        # features which must be redirected on rooting.
        testnode = self.tree.treenode.get_leaves()[0]
        extrafeat = {i for i in testnode.features if i not in self.features}
        self.features.update(extrafeat)



    def get_match(self, names, wildcard, regex):
        # find the selected node
        self.nas = NodeAssist(self.tree, names, wildcard, regex)
        self.nas.match_query()
        self.tipnames = self.nas.get_tipnames()

        # check for reciprocal match
        x0 = (not self.nas.is_query_monophyletic())
        x1 = (self.nas.get_mrca().is_root())
        if x0 or x1:
            clade1 = self.nas.tipnames
            self.nas.match_reciprocal()
            if not self.nas.is_query_monophyletic():
                clade2 = self.nas.tipnames
                raise ToytreeError(
                    "Matched query is paraphyletic: {}"
                    .format(sorted([clade1, clade2], key=len)[0]))



#!/usr/bin/env python

from __future__ import print_function, division, absolute_import


import re
from copy import deepcopy
import numpy as np


#######################################################
# Exception Classes
#######################################################
class ToytreeError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class TreeError(Exception):
    "A problem occurred during a TreeNode operation"
    def __init__(self, value=''):
        self.value = value

    def __str__(self):
        return repr(self.value)


# TREE FORMATS
NW_FORMAT = {
    # flexible with support
    # Format 0 = (A:0.35,(B:0.72,(D:0.60,G:0.12)1.00:0.64)1.00:0.56);
    0: [
        ('name', str, True),
        ('dist', float, True),
        ('support', float, True),
        ('dist', float, True),
    ],

    # flexible with internal node names
    # Format 1 = (A:0.35,(B:0.72,(D:0.60,G:0.12)E:0.64)C:0.56);
    1: [
        ('name', str, True),
        ('dist', float, True),
        ('name', str, True),
        ('dist', float, True),      
    ],

    # strict with support values
    # Format 2 = (A:0.35,(B:0.72,(D:0.60,G:0.12)1.00:0.64)1.00:0.56);
    2: [
        ('name', str, False),
        ('dist', float, False),
        ('support', str, False),
        ('dist', float, False),      
    ],

    # strict with internal node names
    # Format 3 = (A:0.35,(B:0.72,(D:0.60,G:0.12)E:0.64)C:0.56);
    3: [
        ('name', str, False),
        ('dist', float, False),
        ('name', str, False),
        ('dist', float, False),      
    ],

    # strict with internal node names
    # Format 4 = (A:0.35,(B:0.72,(D:0.60,G:0.12)));
    4: [
        ('name', str, False),
        ('dist', float, False),
        (None, None, False),
        (None, None, False),      
    ],

    # Format 5 = (A:0.35,(B:0.72,(D:0.60,G:0.12):0.64):0.56);
    5: [
        ('name', str, False),
        ('dist', float, False),
        (None, None, False),
        ('dist', float, False),      
    ],

    # Format 6 = (A:0.35,(B:0.72,(D:0.60,G:0.12)E)C);
    6: [
        ('name', str, False),
        (None, None, False),
        (None, None, False),        
        ('dist', float, False),      
    ],

    # Format 7 = (A,(B,(D,G)E)C);
    7: [
        ('name', str, False),
        ('dist', float, False),
        ('name', str, False),        
        (None, None, False),        
    ],    


    # Format 8 = (A,(B,(D,G)));
    8: [
        ('name', str, False),
        (None, None, False),
        ('name', str, False),        
        (None, None, False),
    ],

    # Format 9 = (,(,(,)));
    9: [
        ('name', str, False),
        (None, None, False),
        (None, None, False),
        (None, None, False),
    ],    

    # Format 10 = ((a[&Z=1,Y=2]:1.0[&X=3], b[&Z=1,Y=2]:3.0[&X=2]):1.0[&L=1,W=0], ...
    # NHX Like mrbayes NEXUS common
    10: [
        ('name', str, True),
        ('dist', str, True),
        ('name', str, True),
        ('dist', str, True),
    ]
}






# class TreeInference:
# - get distance matrix (from an input data set... phy, nex)
# - ----- create a class to store DNA matrix (pandas colored)
# - NJ tree infer
#   ------ uses distance matrix
# - UPGMA tree infer
#   ------ uses distance matrix


#class TreeMoves:
#     def move_spr(self):
#         """
#         Sub-tree pruning and Regrafting. 
#         Select one edge randomly from the tree and split on that edge to create
#         two subtrees. Attach one of the subtrees (e.g., the smaller one) 
#         randomly to the larger tree to create a new node.
#         ... does SPR break edges connected to root when tree is real rooted?
#         """
#         pass
#         # On rooted trees we can work with nodes easier than edges. Start by
#         # selected a node at random that is not root.
#         # nodes = [i for i in self.ttree.tree.traverse() if not i.is_root()]
#         # rnode = nodes[random.randint(0, len(nodes) - 1)]
#         # # get all edges on the tree, skip last one which is non-real root edge
#         # edges = self.ttree.tree.get_edges()[:-1]
#         # # select a random edge
#         # redge = edges[random.randint(0, len(edges))]
#         # # break into subtrees
#         # tre1 = self.tree.prune(self.tree.get_common_ancestor(redge[0]).idx)
#         # tre2 = self.tree.prune(self.tree.get_common_ancestor(redge[1]).idx)



#     def move_tbr(self):
#         pass


#     def move_nni(self):
#         pass


#     def non_parametric_rate_smoothing(self):
#         """
#         Non-parametric rate smooting.
#         A method for estimating divergence times when evolutionary rates are 
#         variable across lineages by minimizing ancestor-descendant local rate
#         changes. According to Sanderson this method is motivated by the 
#         likelihood that evolutionary rates are autocorrelated in time.

#         returns Toytree
#         """
#         # p is a fixed exponent
#         p = 2
#         W = []
#         for node in self.ttree.traverse():
#             if not node.is_leaf():
#                 children = node.children
#                 ks = []
#                 for child in children:
#                     dist = abs(node.dist - child.dist)
#                     ks.append(dist ** p)
#                 W.append(sum(ks))

#         # root rate is mean of all descendant rates -- 
#         # n is the number of edges (rates) (nnodes - 1 for root)
#         r_root = np.mean(W)
#         rootw = []
#         for child in self.ttree.tree.children:
#             rootw.append((r_rroot - child.dist) ** p)
#         w_root = sum(rootw)
#         W.append(w_root)
#         k = []
#         for 
#         k = sum(  np.exp(abs(ri - rj), p)  )
#         W = sum(k)


#     def penalized_likelihood(...):
#         pass
#

# def wfunc(ttree, p):
#     ws = []
#     for node in ttree.tree.traverse():
#         if not node.is_leaf():          
#             w = sum([(node.dist - child.dist) ** p for child in node.children])
#             ws.append(w)
#     return sum(ws)






#######################################################
# Other
#######################################################
def bpp2newick(bppnewick):
    "converts bpp newick format to normal newick"
    regex1 = re.compile(r" #[-+]?[0-9]*\.?[0-9]*[:]")
    regex2 = re.compile(r" #[-+]?[0-9]*\.?[0-9]*[;]")
    regex3 = re.compile(r": ")
    new = regex1.sub(":", bppnewick)
    new = regex2.sub(";", new)
    new = regex3.sub(":", new)
    return new



# TODO: would be useful for (eg., root) to have option to return not mrca, 
# and fuzzy match just tips, or nodes, etc...




def normalize_values(vals, nbins=10, minsize=2, maxsize=12):
    """
    Distributes values into bins spaced at reasonable sizes for plotting.
    Example, this can be used automatically scale Ne values to plot as 
    edge widths.
    """

    # make copy of original
    ovals = deepcopy(vals)

    # if 6X min value is higher than max then add this 
    # as a fake value to scale more nicely
    vals = list(vals)
    if min(vals) * 6 > max(vals):
        vals.append(min(vals) * 6)

    # sorted vals list
    svals = sorted(vals)

    # put vals into bins
    bins = np.histogram(vals, bins=nbins)[0]

    # convert binned vals to widths in 2-12
    newvals = {}
    sizes = np.linspace(minsize, maxsize, nbins)
    for idx, inbin in enumerate(bins):
        for num in range(inbin):
            newvals[svals.pop(0)] = sizes[idx]
    return np.array([newvals[i] for i in ovals])



# def fuzzy_match_tipnames(ttree, names, wildcard, regex, mono=True, retnode=True):
def fuzzy_match_tipnames(ttree, names, wildcard, regex, mrca=True, mono=True):
    """
    Used in multiple internal functions (e.g., .root()) and .drop_tips())
    to select an internal mrca node, or multiple tipnames, using fuzzy matching
    so that every name does not need to be written out by hand.

    name: verbose list
    wildcard: matching unique string
    regex: regex expression
    mrca: return mrca node of selected tipnames. 
    mono: raise error if selected tipnames are not monophyletic    
    """
    # require arguments
    if not any([names, wildcard, regex]):
        raise ToytreeError(
            "must enter an outgroup, wildcard selector, or regex pattern")

    # get list of **nodes** from {list, wildcard, or regex}
    tips = []
    if names:
        if isinstance(names, (str, int)):
            names = [names]
        notfound = [i for i in names if i not in ttree.get_tip_labels()]
        if any(notfound):
            raise ToytreeError(
                "Sample {} is not in the tree".format(notfound))
        tips = [i for i in ttree.treenode.get_leaves() if i.name in names]

    # use regex to match tipnames
    elif regex:
        tips = [
            i for i in ttree.treenode.get_leaves() if re.match(regex, i.name)
        ]               
        if not any(tips):
            raise ToytreeError("No Samples matched the regular expression")

    # use wildcard substring matching
    elif wildcard:
        tips = [i for i in ttree.treenode.get_leaves() if wildcard in i.name]
        if not any(tips):
            raise ToytreeError("No Samples matched the wildcard")

    # build list of **tipnames** from matched nodes
    if not tips:
        raise ToytreeError("no matching tipnames")       
    tipnames = [i.name for i in tips]

    # if a single tipname matched no need to check for monophyly
    if len(tips) == 1:
        if mrca:
            return tips[0]
        else:
            return tipnames

    # if multiple nodes matched, check if they're monophyletic
    mbool, mtype, mnames = (
        ttree.treenode.check_monophyly(
            tipnames, "name", ignore_missing=True)
    )

    # get mrca node
    node = ttree.treenode.get_common_ancestor(tips)

    # raise an error if required to be monophyletic but not
    if mono:
        if not mbool:
            raise ToytreeError(
                "Taxon list cannot be paraphyletic")

    # return tips or nodes
    if not mrca:
        return tipnames
    else:
        return node



#!/usr/bin/env python

"""
A class object for generating and storing Toytree plotting coordinates.
"""

import numpy as np
# from .utils import ToyTreeError


class Coords:
    """
    Generates and stores plotting coordinates for nodes and edges of a tree. 
    Uses the toytree _style information (e.g., layout, use_edge_lengths).

    Attributes: 
    -----------
    edges: ndarray
        2-d array of pairs showing (parent, child) relationships.
        Rows are ordered in postorder traversal order for plotting.
    verts: ndarray
        2-d array with relative cartesian coordinate positions of nodes (n)
        Rows are ordered by the numbered 'idx' feature of nodes.
    coords: ndarray
        2-d array with cartesian coordinates for tips and node crown positions.
        Rows are ordered postorder idx.
        Used for drawing edges on phylograms, not for nodes.
    lines: ndarray
        ...
    """
    def __init__(self, ttree):

        # the toytree 
        self.ttree = ttree

        # the vertices and edge tuples for normal layouts ('n')
        self.edges = np.zeros((self.ttree.nnodes - 1, 2), dtype=int)
        self.verts = np.zeros((self.ttree.nnodes, 2), dtype=float)
        self.lines = []
        self.coords = []

        # the class object for transforming to radial coords ('r').
        # init'ing this is pretty lightweight, so might as well default it.
        self.circ = Circle(self.ttree)

        # the class object for transforming to force-directed layout ('u') 
        # ...

        # not automatically run in init bc we want to keep toytrees light
        # self.update()


    # it may be useful to separate these so we can sometimes keep idxs static
    # while only changing the plotting coordinates...
    def update(self):
        "Updates cartesian coordinates for drawing tree graph"              
        # get new shape and clear for attrs
        self.edges = np.zeros((self.ttree.nnodes - 1, 2), dtype=int)
        self.verts = np.zeros((self.ttree.nnodes, 2), dtype=float)
        self.lines = []
        self.coords = []
        self.circ = Circle(self.ttree)

        # updates idxs and fixed_idx for any tree manipulations
        self.update_idxs()             # get dimensions of tree
        self.update_fixed_order()      # in case ntips changed

        # get edges and verts (node locations)
        if self.ttree.style.layout in ("c", "circ", "circular", "x", "unrooted"):
            self.assign_radial_vertices()

        else:           
            self.assign_vertices()

        # if unrooted use these vertices as start in a force directed push
        # if self.ttree.style.layout in ("x", "unrooted"):
        # self.force_directed_verts()

        # get lines and coords (node neighbor locations for 'p' edges)
        # self.new_assign_coordinates()      # get edge locations
        self.assign_coordinates()      # get edge locations        
        self.reorient_coordinates()    # orientation can reorder dimensions


    # # NOT YET IMPLEMENTED
    # def force_directed_verts(self):
    #     # pin 25% of 
    #     c, a, m = toyplot.graph(
    #         self.coords.edges,
    #         layout=toyplot.layout.FruchtermanReingold(
    #             seed=self._seed, 
    #             area=100, 
    #             M=self.ttree.ntips*20, 
    #             temperature=10,
    #             ),
    #     )
    #     # store the coordinates after the push
    #     # self._unrooted_coords = np.array(m.vcoordinates)
    #     # get circular lines and coords
    #     self.assign_coordinates()      # get edge locations
    #     self.reorient_coordinates()    # orientation can reorder dimensions


    def update_idxs(self):
        "set root idx highest, tip idxs lowest ordered as ladderized"
        # internal nodes: root is highest idx
        idx = self.ttree.nnodes - 1
        for node in self.ttree.treenode.traverse("levelorder"):
            if not node.is_leaf():
                node.add_feature("idx", idx)
                if not node.name:
                    node.name = str(idx)
                idx -= 1

        # external nodes: lowest numbers are for tips (0-N)
        for node in self.ttree.treenode.get_leaves():
            node.add_feature("idx", idx)
            if not node.name:
                node.name = str(idx)
            idx -= 1


    def update_fixed_order(self):
        "after pruning fixed order needs update to match new nnodes/ntips."
        # set tips order if fixing for multi-tree plotting (default None)
        fixed_order = self.ttree._fixed_order
        self.ttree_fixed_order = None
        self.ttree_fixed_idx = list(range(self.ttree.ntips))

        # check if fixed_order changed:
        if fixed_order:
            fixed_order = [
                i for i in fixed_order if i in self.ttree.get_tip_labels()]
            self.ttree._set_fixed_order(fixed_order)
        else:
            self.ttree._fixed_idx = list(range(self.ttree.ntips))


    def assign_radial_vertices(self):
        """
        Assign .edges and .verts for node positions in a fan tree.
        The farthest tip aligns at the circumference.
        """
        # shortname 
        uselen = bool(self.ttree.style.use_edge_lengths)

        # store edge array for connecting child nodes to parent nodes
        nidx = 0
        for node in self.ttree.treenode.traverse("postorder"):            
            if not node.is_root():
                self.edges[nidx, :] = [node.up.idx, node.idx]
                nidx += 1

        # used for fixed-order setting
        # tidx = len(self.ttree) - 1

        # store verts 
        for node in self.ttree.treenode.traverse("postorder"):

            # leaves: x positions are evenly spaced around circumference
            if node.is_leaf() and (not node.is_root()):

                # get positions of tips using radians and radius
                node.radians = self.circ.tip_radians[node.idx]
                if uselen:
                    node.radius = self.circ.radius - node.height
                    node.x, node.y = self.circ.get_node_coords(node)
                else:
                    node.radius = self.circ.radius
                    node.x, node.y = self.circ.get_node_coords(node)

                # TODO: allow fixed order in fan
                # if self.ttree._fixed_order:
                    # node.x = self.ttree._fixed_order.index(node.name)
                # else:
                    # node.x = tidx
                    # tidx -= 1
                # node.x = circ.circ[node.x][0]
                self.verts[node.idx] = [node.x, node.y]

            # internal nodes comes after tips. Inherit position from children.
            else:

                # height is either distance or nodes from root
                if uselen:
                    node.radius = self.circ.radius - node.height
                else:
                    node.radius = max([i.radius for i in node.children]) - 1

                # x position is halfway between children x-pos
                node.radians = np.mean([i.radians for i in node.children])
                node.x, node.y = self.circ.get_node_coords(node)
                # if node.children:
                    # node.x = circ.get_circumference_midpoint(*node.children)
                # else:
                    # node.x = circ.circ[tidx]

                # store the x,y vertex positions
                self.verts[node.idx] = [node.x, node.y]


    def assign_vertices(self):
        """
        Sets .edges, .verts for node positions. 
        X and Y positions here refer to base assumption that tree is right
        facing, reorient_coordinates() will handle re-translating this.        
        """
        # shortname 
        uselen = bool(self.ttree.style.use_edge_lengths)

        # postorder: children then parents (nidxs from 0 up)
        # store edge array for connecting child nodes to parent nodes
        nidx = 0
        for node in self.ttree.treenode.traverse("postorder"):            
            if not node.is_root():
                self.edges[nidx, :] = [node.up.idx, node.idx]
                nidx += 1

        # store verts array with x,y positions of nodes (lengths of branches)
        # we want tips to align at the right face (larger axis number)
        _root = self.ttree.treenode.get_tree_root()
        _treeheight = _root.get_distance(_root.get_farthest_leaf()[0])

        # set node x, y
        tidx = len(self.ttree) - 1
        for node in self.ttree.treenode.traverse("postorder"):

            # Just leaves: x positions are evenly spread and ordered on axis
            if node.is_leaf() and (not node.is_root()):

                # set y-positions (heights). Distance from root or zero
                node.y = _treeheight - _root.get_distance(node)
                if not uselen:
                    node.y = 0.0

                # set x-positions (order of samples)
                if self.ttree._fixed_order:
                    node.x = self.ttree._fixed_order.index(node.name)# - tidx
                else:
                    node.x = tidx
                    tidx -= 1

                # store the x,y vertex positions
                self.verts[node.idx] = [node.x, node.y]

            # All internal node positions are not evenly spread or ordered
            else:
                # height is either distance or nnodes from root
                node.y = _treeheight - _root.get_distance(node)
                if not uselen:
                    node.y = max([i.y for i in node.children]) + 1

                # x position is halfway between childrens x-positions
                if node.children:
                    nch = node.children
                    node.x = sum(i.x for i in nch) / float(len(nch))
                else:
                    node.x = tidx

                # store the x,y vertex positions                    
                self.verts[node.idx] = [node.x, node.y]


    # IN DEVELOPMENT: 
    # this will recude lines and allow curved eges on circular trees.
    #
    def new_assign_coordinates(self):
        """
        coords span from left most to right most child.
        """
        if len(self.ttree) < 2:
            return 

        # a list for storing edge tuples
        edges = []

        # a dict of all existing nodes indexed by idx, we will add to this.
        coords = {i: tuple(j) for (i, j) in enumerate(self.verts)}

        # idx of the next node that will be created
        nidx = self.ttree.treenode.idx + 1

        # add edges connecting tips up to their bracket edge
        for node in self.ttree.treenode.traverse():
            if not node.is_root():

                # connect node: ##  x-v-n  ################
                # y to x            |   |
                #                   y   z
                #                 
                # get coordinates for the new node nidx          
                if self.ttree.style.layout == 'c':
                    x, y = self.circ.get_node_lines(node)
                else:
                    x, y = (node.x, node.up.y)

                # store coordinates of new node nidx
                coords[nidx] = x, y

                # add connection of new node to its child
                edges.append((nidx, node.idx))

                # store node is the child of nidx
                node.nup = nidx

                # advance node counter
                nidx += 1
            else:
                edges.append((node.idx, node.idx))

        # add side edges connecting bracket arms
        for node in self.ttree.treenode.traverse("levelorder"):
            if len(node.children) > 1:
                # connect node: ##  x-v-n  ################
                # x to n            |   |
                #                   y   z
                #                 
                # get coordinates for the new node nidx
                edges.append((node.children[0].nup, node.children[-1].nup))

        # store coords and lines back into ndarrays
        self.coords = np.array([coords[i] for i in range(len(coords))])
        self.lines = np.array(edges)


    def assign_coordinates(self):
        """
        coords and lines allow 'p' type edges (not all straight lines)
        """
        if len(self.ttree) < 2:
            return 

        edges = []
        coords = {i: tuple(j) for (i, j) in enumerate(self.verts)}
        nidx = self.ttree.treenode.idx + 1

        # add up nodes and edges
        for node in self.ttree.treenode.traverse():
            if not node.is_root():
                if self.ttree.style.layout == 'c':
                    coords[nidx] = self.circ.get_node_lines(node)
                else:
                    coords[nidx] = (node.x, node.up.y)
                edges.append((nidx, node.idx))
                node.nup = nidx
                nidx += 1

        # add side edges
        for node in self.ttree.treenode.traverse():
            if not node.is_leaf():
                for child in node.children:
                    edges.append((node.idx, child.nup))

        # store the vertex coordinates as an array
        self.coords = np.array([coords[i] for i in range(len(coords))])

        # store the edges as an array
        self.lines = np.array(edges)


    def reorient_coordinates(self):
        """
        Returns a modified .verts array with new coordinates for nodes. 
        This does not need to modify .edges. The order of nodes, and therefore
        of verts rows is still the same because it is still based on the tree
        branching order (ladderized usually). 
        """
        # if tree is empty then bail out
        if len(self.ttree) < 2:
            return

        # TODO: orientation for non linear trees
        if self.ttree.style.layout in ('c', 'circular', 'x', 'unrooted'):
            return 

        # default: Down-facing tips align at y=0, first ladderized tip at x=0
        if self.ttree.style.layout in ('d', 'down'):
            pass

        # right-facing tips align at x=0, last ladderized tip at y=0
        elif self.ttree.style.layout in ('r', 'right'):

            # verts swap x and ys and make xs 0 to negative
            tmp = np.zeros(self.verts.shape)
            tmp[:, 1] = self.verts[:, 0]
            tmp[:, 0] = self.verts[:, 1] * -1
            self.verts = tmp

            # coords...
            tmp = np.zeros(self.coords.shape)
            tmp[:, 1] = self.coords[:, 0]
            tmp[:, 0] = self.coords[:, 1] * -1
            self.coords = tmp

        elif self.ttree.style.layout in ('l', 'left'):

            # verts swap x and ys and make xs 0 to negative
            tmp = np.zeros(self.verts.shape)
            tmp[:, 1] = self.verts[:, 0]
            tmp[:, 0] = self.verts[:, 1] 
            self.verts = tmp

            # coords...
            tmp = np.zeros(self.coords.shape)
            tmp[:, 1] = self.coords[:, 0]
            tmp[:, 0] = self.coords[:, 1]
            self.coords = tmp

        elif self.ttree.style.layout in ('u', 'up'):
            # verts swap x and ys and make xs 0 to negative
            tmp = np.zeros(self.verts.shape)
            tmp[:, 1] = self.verts[:, 1] * -1
            tmp[:, 0] = self.verts[:, 0] * -1
            self.verts = tmp

            # coords...
            tmp = np.zeros(self.coords.shape)
            tmp[:, 1] = self.coords[:, 1] * -1
            tmp[:, 0] = self.coords[:, 0] * -1
            self.coords = tmp

        else:
            raise ToyTreeError("layout not recognized")



class Circle:
    """
    When init from a toytree it can return coordinates based on 
    distributing tips equally in radians from 0-2pi (TODO: allow 
    setting start/end).

    The radius for farthest tip-ends is set to the tree height, 
    and origin is at 0.0.
    """
    def __init__(self, tre):

        # set radius
        self.tre = tre
        self.radius = self.tre.treenode.height 
        # get_distance(self.tre.treenode.get_farthest_leaf()[0])

        # origin
        self.o = (0, 0)
        # self.tre.style.xbaseline, self.tre.style.ybaseline)  # -self.radius, 0)

        # tips (bottom to top) are evenly spread from 0 to -2pi (counter clock)
        self.tip_radians = np.linspace(0, -np.pi * 2, self.tre.ntips + 1)[:-1]


    def get_node_coords(self, node):
        """
        get node coord. Node has radius assigned on the fly.
        """
        x = self.o[0] + node.radius * np.cos(node.radians)
        y = self.o[1] - node.radius * np.sin(node.radians)
        return x, y


    def get_node_lines(self, node):
        x = self.o[0] + node.up.radius * np.cos(node.radians)
        y = self.o[1] - node.up.radius * np.sin(node.radians)
        return x, y      


    def get_tip_end_angles(self):
        """
        node radians for printing tip labels should be from the root (0,0)
        """
        return np.array([np.rad2deg(abs(i)) for i in self.tip_radians])


    def get_tip_end_coords(self):
        """
        node tip coords must calculate new radian angle relative to parent 
        node and then add offset amount to radius when calculating (x, y).
        """
        xs = [self.o[0] + self.radius * np.cos(i) for i in self.tip_radians]        
        ys = [self.o[1] - self.radius * np.sin(i) for i in self.tip_radians]
        return np.stack([xs, ys], axis=1)


    # def get_parent_coords(self, node):
    #     """
    #     get x,y coordinates of parent based on radians of children
    #     """
    #     # calculate x, y coords
    #     x = self.o[0] + node.radius * np.cos(midradian)
    #     y = self.o[1] + node.radius * np.sin(midradian)
    #     return x, y




#!/usr/bin/env python

__version__ = "1.1.4"
__author__ = "Deren Eaton"

from .Toytree import ToyTree as tree
from .Toytree import RawTree as _rawtree
from .Randomtree import RandomTree as rtree
from .Multitree import MultiTree as mtree
from .Container import Container as container
from .PCM import PCM as pcm

# make a color palette easily accessible and an iter cycling version
from .TreeStyle import COLORS1 as colors
from .TreeStyle import COLORS2 as darkcolors
import itertools as _itertools
icolors1 = _itertools.cycle(colors)
icolors2 = _itertools.cycle(darkcolors)



#!/usr/bin/env python

"""
Draw admixture edges on a tree
"""

import itertools
import numpy as np
from .TreeStyle import COLORS1


class AdmixEdges():
    """
    TODO...

    Parameters
    ----------
    tree: toytree

    tupleargs: list
        A list of admixture tuples.

    layout: int
        0 is 'down' layout and 1 is 'right' layout.
    """
    def __init__(
        self,
        tree,
        axes,
        admixture_edges,
        layout="d",
        edge_type="c",
        ):
        # source, dest, sheight, dheight, offset, layout):

        # store args
        self.tree = tree
        self.axes = axes
        self.tupleargs = admixture_edges

        # edge-type layout
        self.edge_type = bool(edge_type == "c")

        # orientation layout
        self.layout = bool(layout == "d")

        # use a cycling color palette
        self.colors = itertools.cycle(COLORS1)
        next(self.colors)

        # parse tuples into ...
        if self.tupleargs is not None:

            # check and fill tuples
            self.check_tuple_args()

            # draw edges as a toyplot graph on the axes
            self.draw()



    def check_tuple_args(self):
        """
        Split list into bits.
        """
        # check that the argument is a list
        assert isinstance(self.tupleargs, list), (
            "admixture edges should be a list of tuples")

        # iterate over the tuples
        self.admixture_edges = []
        for tup in self.tupleargs:

            # parse required args
            source = tup[0]
            dest = tup[1]

            # parse optional args
            try:
                height = tup[2]
            except IndexError:
                height = None

            try:
                width = (tup[3] if tup[3] else 3)
            except IndexError:
                width = 3

            try:
                color = (tup[4] if tup[4] else next(self.colors))
            except IndexError:
                color = next(self.colors)

            try:
                offset = (tup[5] if tup[5] else 0.05)
            except IndexError:
                offset = 0.05

            # put it back into list as full tuples
            self.admixture_edges.append(
                (source, dest, height, width, color, offset))



    def draw(self):

        # iterate over edge tuples
        for aedge in self.admixture_edges:

            # get vcoordinates for this edge
            vcoords = get_admix_verts(
                self.tree, 
                aedge[0], 
                aedge[1],
                aedge[2],
                aedge[5],
                self.edge_type,
                self.layout,
            )

            # generate marker on axes
            self.axes.graph(
                np.array([[0, 1], [1, 2], [2, 3]]),
                vcoordinates=vcoords,
                vsize=0,
                vlshow=False,
                eopacity=1,
                ewidth=aedge[3],
                ecolor=aedge[4],
                estyle={"stroke-linecap": "round"},
                # mmarker=[None, ">", None]
            )
        return self.axes




def get_admix_verts(tree, source, dest, height, offset, etype, layout):
    """
    Get coordinates for graph edges.
    """
    # get node dictionary
    ndict = tree.get_node_dict(True, True)

    # use midpoint edge if no heights given
    if height is None:
        edgedict = get_all_admix_edges(tree, 0.5, 0.5)
        height = np.mean(edgedict[(source, dest)])

    # get the child vertex coordinates that source starts from
    cy, cx = tree._coords.verts[source]
    py, px = tree._coords.verts[ndict[source].up.idx]
    a, b = cx, cy

    # get the vertex on the edge above source at (x, height)
    if etype:
        theta = np.arctan((px - cx) / (py - cy))
        bx = (np.sin(theta) * (height + cy)) - cx
        by = np.cos(theta) * height
    else:
        bx = -cx
        by = height
    c, d = -bx, -by

    # get the vertex of the parent node of destination
    cy, cx = tree._coords.verts[dest]
    py, px = tree._coords.verts[ndict[dest].up.idx]
    if etype:
        g, h = px, py
    else:
        g, h = cx, py

    # get the vertex on the edge of destination
    if etype:
        theta = np.arctan((px - cx) / (py - cy))
        bx = (np.sin(theta) * (height + cy)) - cx
        by = np.cos(theta) * height
        e, f = -bx, -by
    else:
        e, f, = cx, -height

    # order in final array
    arr = np.array([
        [a + offset, -b],
        [c + offset, -d],
        [e + offset, -f],
        [g + offset, -h],
    ])

    # how is the tree oriented
    if not layout:
        arr = arr[:, ::-1]
        arr[:, 0] = arr[:, 0] * -1

    return arr




def get_all_admix_edges(ttree, lower=0.25, upper=0.75, exclude_sisters=False):
    """
    Find all possible admixture edges on a tree. 

    Edges are unidirectional, so the source and dest need to overlap in
    time interval. To retrict migration to occur away from nodes (these 
    can be harder to detect when validating methods) you can set upper 
    and lower limits. For example, to make all source migrations to occur
    at the midpoint of overlapping intervals in which migration can occur
    you can set upper=.5, lower=.5.   
    """
    # bounds on edge overlaps
    if lower is None:
        lower = 0.0
    if upper is None:
        upper = 1.0

    # for all nodes map the potential admixture interval
    for snode in ttree.treenode.traverse():
        if snode.is_root():
            snode.interval = (None, None)
        else:
            snode.interval = (snode.height, snode.up.height)

    # for all nodes find overlapping intervals
    intervals = {}
    for snode in ttree.treenode.traverse():
        for dnode in ttree.treenode.traverse():
            if not any([snode.is_root(), dnode.is_root(), dnode == snode]):

                # [option] skip sisters
                if (exclude_sisters) & (dnode.up == snode.up):
                    continue

                # check for overlap
                smin, smax = snode.interval
                dmin, dmax = dnode.interval

                # find if nodes have interval where admixture can occur
                low_bin = np.max([smin, dmin])
                top_bin = np.min([smax, dmax])              
                if top_bin > low_bin:

                    # restrict migration within bin to a smaller interval
                    length = top_bin - low_bin
                    low_limit = low_bin + (length * lower)
                    top_limit = low_bin + (length * upper)
                    intervals[(snode.idx, dnode.idx)] = (low_limit, top_limit)
    return intervals



#!/usr/bin/env python

"""
PCM: phylogenetic comparative methods tools
"""

import numpy as np


class PCM:
    """
    Phylogenetic comparative methods implemented on toytrees.
    """
    def __init__(self, tree):
        self.tree = tree


    def independent_contrasts(self, feature):
        ntree = self.tree.copy()
        resdict = PIC(ntree, feature)
        ntree = ntree.set_node_values(
            feature="{}-contrast",
            values={i.name: j[2] for (i, j) in resdict.items()}
        )
        ntree = ntree.set_node_values(
            feature="{}-contrast-var",
            values={i.name: j[3] for (i, j) in resdict.items()}
        )        
        return ntree


    def ancestral_state_reconstruction(self, feature):
        """
        Infer ancestral states on ancestral nodes for continuous traits
        under a brownian motion model of evolution.

        Modified from IVY interactive (https://github.com/rhr/ivy/)

        Returns a toytree with feature applied to each node.
        """
        ntree = self.tree.copy()
        resdict = PIC(ntree, feature)
        ntree = ntree.set_node_values(
            feature, 
            values={i.name: j[0] for (i, j) in resdict.items()}
        )
        return ntree


    def tree_to_VCV(self):
        return VCV(self.tree)



def VCV(tree):
    """
    Return the variance co-variance metrix representing the tree topology.
    """
    vcv_ = np.zeros((tree.ntips,tree.ntips))
    labs = tree.get_tip_labels()
    for lab1 in range(tree.ntips):
        for lab2 in range(tree.ntips):
            mrca_idx = tree.get_mrca_idx_from_tip_labels([labs[lab1],labs[lab2]])
            mrca_height = tree.treenode.search_nodes(idx=mrca_idx)[0].height
            vcv_[lab1, lab2] = tree.treenode.height - mrca_height
    return(vcv_)



def PIC(tree, feature):
    """
    Infer ancestral states and calculate phylogenetic independent
    contrasts at nodes for a selected feature (trait). 

    Modified from IVY interactive (https://github.com/rhr/ivy/)

    Parameters
    ----------
    feature: (str)
        The name of a feature of the tree that has been mapped to all 
        tip nodes of the tree. 

    Returns
    -------
    toytree (Toytree.ToyTree)
        A modified copy of the input tree is returned with the mean ancestral
        value for the selected feature inferred for all nodes of the tree. 
    """
    # get current node features at the tips
    fdict = tree.get_feature_dict(key_attr="name", values_attr=feature)
    data = {i: j for (i, j) in fdict.items() if i in tree.get_tip_labels()}

    # apply dynamic function from ivy to return dict results
    results = dynamicPIC(tree.treenode, data, results={})

    # return dictionary mapping nodes to (mean, var, contrast, cvar)
    return results


def dynamicPIC(node, data, results):
    """
    Phylogenetic independent contrasts. Recursively calculate 
    independent contrasts of a bifurcating node given a dictionary
    of trait values.

    Modified from IVY interactive (https://github.com/rhr/ivy/)

    Args:
        node (Node): A node object
        data (dict): Mapping of leaf names to character values
    Returns:
        dict: Mapping of internal nodes to tuples containing ancestral
              state, its variance (error), the contrast, and the
              contrasts's variance.
    TODO: modify to accommodate polytomies.
    """    
    X = []
    v = []

    # recursively does children until X and v are full
    for child in node.children:

        # child has children, do those first
        if child.children:

            # update results dict with children values
            dynamicPIC(child, data, results)
            child_results = results[child]

            # store childrens values
            X.append(child_results[0])
            v.append(child_results[1])

        # no child of child, so just do child
        else:
            X.append(data[child.name])
            v.append(child.dist)

    # Xi - Xj is the contrast value
    Xi, Xj = X  

    # vi + vj is the contrast variance
    vi, vj = v

    # Xk is the reconstructed state at the node
    Xk = ((1.0 / vi) * Xi + (1 / vj) * Xj) / (1.0 / vi + 1.0 / vj)

    # vk is the variance
    vk = node.dist + (vi * vj) / (vi + vj)

    # store in dictionary and 
    results[node] = (Xk, vk, Xi - Xj, vi + vj)
    return results


# single test
if __name__ == "__main__":

    import toyplot
    import toytree

    colormap = toyplot.color.brewer.map("BlueRed", reverse=True)
    colormap

    tree = toytree.rtree.imbtree(5, 1e6)
    tree = tree.set_node_values(
        "g", 
        values={i: 5 for i in (2, 3, 4)},
        default=1,
    )
    tree.draw(
        ts='p', 
        node_labels=tree.get_node_values("g", 1, 1),
        node_colors=[
            colormap.colors(i, 0, 5) for i in tree.get_node_values('g', 1, 1)]
        )

    # apply reconstruction
    ntree = PIC(tree, "g")

    # new values are stored as -mean, -var, -contrasts, ...
    evals = ntree.get_edge_values("g-mean")

    # draw new tree
    ntree.draw(
        ts='p', 
        node_labels=ntree.get_node_values("g-mean", 1, 1),
        node_colors=[
            colormap.colors(i, 0, 5) for i in 
            ntree.get_node_values('g-mean', 1, 1)]
    )



#!/usr/bin/env python


import re
from .utils import ToytreeError



class NodeAssist:
    """
    Given a search query (list of names, wildcard or regex) a node or list of 
    names can be retrieved under a set of pre-built functions.
    """
    def __init__(self, ttree, names, wildcard, regex):

        # attributes
        self.ttree = ttree
        self.names = names
        self.wildcard = wildcard
        self.regex = regex

        # require arguments
        if not any([names, wildcard, regex]):
            raise ToytreeError(
                "Must enter a name list, wildcard selector, or regex pattern")

        if len([i for i in [names, wildcard, regex] if i]) != 1:
            raise ToytreeError(
                "Only one method allowed at a time for: name list, "
                "wildcard selector, or regex pattern")

        # matched values
        self.nodes = []
        self.tipnames = []
        self.match_query()

        # default options to be updated in function calls
        self.mrca = True
        self.monophyletic = True


    def match_query(self):
        """
        Get list of **nodes** from {list, wildcard, or regex}
        """
        tips = []
        if self.names:

            # allow tips to be entered instead of a list
            if isinstance(self.names, (str, int)):
                self.names = [self.names]

            # report any names entered that seem like typos
            bad = [i for i in self.names if i not in self.ttree.get_tip_labels()]
            if any(bad):
                raise ToytreeError(
                    "Sample {} is not in the tree".format(bad))

            # select *nodes* that match these names
            tips = [
                i for i in self.ttree.treenode.get_leaves() 
                if i.name in self.names
            ]

        # use regex to match tipnames
        elif self.regex:

            # select *nodes* that regex match. Raise error if None.
            tips = [
                i for i in self.ttree.treenode.get_leaves() 
                if re.match(self.regex, i.name)
            ]               
            if not any(tips):
                raise ToytreeError("No Samples matched the regular expression")

        # use wildcard substring matching
        elif self.wildcard:

            # select *nodes* that match the wildcard search
            tips = [
                i for i in self.ttree.treenode.get_leaves()
                if self.wildcard in i.name
            ]
            if not any(tips):
                raise ToytreeError("No Samples matched the wildcard")

        # build list of **tipnames** from matched nodes
        if not tips:
            raise ToytreeError("no matching tipnames")

        # store results
        self.nodes = tips
        self.tipnames = [i.name for i in tips]


    def match_reciprocal(self):
        # get new query names list
        alltips = set(self.ttree.get_tip_labels())
        query_tips = set(self.tipnames)
        other_tips = list(alltips - query_tips)

        # rerun query matching
        self.names = other_tips
        self.wildcard = None
        self.regex = None
        self.match_query()


    def get_nodes(self):
        return self.nodes

    def get_tipnames(self):
        return self.tipnames

    def get_mrca(self):
        if len(self.nodes) > 1:
            return self.ttree.treenode.get_common_ancestor(self.nodes)
        return self.nodes[0]

    def get_mrca_leaves(self):
        return self.get_mrca().get_leaves()

    def get_mrca_descendants(self):
        return [i for i in self.get_mrca().traverse()]

    def is_query_monophyletic(self):
        # if multiple tips descendant from check if they're monophyletic
        if len(self.tipnames) < 2:
            return True

        mbool, mtype, mnames = (
            self.ttree.treenode.check_monophyly(
                self.tipnames, "name", ignore_missing=True)
        )
        return mbool



#!/usr/bin/env python

"""
A class creating Drawings from Toytrees.
"""
from copy import deepcopy
from decimal import Decimal
import numpy as np
import toyplot

from .Admixture import AdmixEdges
from .utils import ToytreeError


# for setting values from iterables
ITERABLE = (list, tuple, np.ndarray)

# should we store node_labels, node_sizes, etc here or in Style?
# It could be that style is just the rules for filling the drawing attrs...
# and the actual drawing features are storing in ._drawing. 


class Drawing:
    def __init__(self, ttree, **kwargs):
        # input objects
        self._tree = ttree.copy()
        self.ttree = ttree
        self.coords = ttree._coords
        self.style = ttree.style
        self.kwargs = kwargs
        self.nedges = self.ttree._coords.edges.shape[0]

        # mutable plotting attributes pulled from styles and tree
        self.node_labels = [""] * self.ttree.nnodes
        self.node_colors = [None] * self.ttree.nnodes
        self.node_sizes = [0] * self.ttree.nnodes
        self.node_markers = ["o"] * self.ttree.nnodes

        self.tip_colors = None
        self.tip_labels = None

        # todo: allow itemized versions of these...
        self.edge_colors = [None] * self.nedges
        self.edge_widths = [None] * self.nedges

        # unrooted layout, allow push layout and store from graph func
        self._unrooted_coords = None
        self._seed = (1234 if "seed" not in self.kwargs else self.kwargs.get("seed"))

        #self.nedges = self.ttree._coords.lines.shape[0]        
        # store whether external axes were passed in 
        self._external_axis = False


    def update(self, axes=None):

        # check layout args
        self.check_layout()

        # always update coords in case style params affect the node placement.
        # this will place nodes for 'n' or 'f' trees, but 'u' trees are auto.
        self.coords.update()

        # set up base canvas and axes, but we need tip labels first
        self.assign_tip_labels_and_colors()
        self.get_dims_from_tree_size()
        self.get_canvas_and_axes(axes)
        self.set_baselines()

        # update attrs for style entries. Some of these can be set with style, 
        # or as a list, e.g., node_style={'fill':'red'} or node_color="red".
        self.assign_node_labels_and_sizes()
        self.assign_node_colors_and_style()
        self.assign_edge_colors_and_widths()

        # draw tree, nodes, tips, axes on canvas.
        self.add_tip_lines_to_axes()
        self.add_tree_to_axes()

        # draw admixture edges 
        self.add_admixture_edges()

        # nodes and tips
        self.add_tip_labels_to_axes()
        self.add_nodes_to_axes()
        self.add_axes_style()


        # add extra display space for tips on the end of tree
        self.fit_tip_labels()
        return self.canvas, self.axes
        #return self.canvas, self.axes, tuple(self.axes._children)


    def check_layout(self):
        if self.style.layout in ('c', 'circ', 'circular'):
            self.style.layout = 'c'
        if self.style.layout in ('x', 'unrooted'):
            self.style.layout = 'x'
        if self.style.layout in ('r', 'right'):
            self.style.layout = 'r'
        if self.style.layout in ('d', 'down'):
            self.style.layout = 'd'
        if self.style.layout in ('l', 'left'):
            self.style.layout = 'l'
        if self.style.layout in ('u', 'up'):
            self.style.layout = 'u'


    def set_baselines(self):
        """
        Modify coords to shift tree position for x,y baseline arguments. This
        is useful for arrangeing trees onto a Canvas with other plots, but 
        still sharing a common cartesian axes coordinates. 
        """
        if self.style.xbaseline:
            self.coords.coords[:, 0] += self.style.xbaseline
            self.coords.verts[:, 0] += self.style.xbaseline                

        if self.style.ybaseline:
            self.coords.coords[:, 1] += self.style.ybaseline
            self.coords.verts[:, 1] += self.style.ybaseline

    # -----------------------------------------------------------------
    # Node and Node Labels 
    # -----------------------------------------------------------------
    def add_tip_labels_to_axes(self):
        """
        Add text offset from tips of tree with correction for orientation, 
        and fixed_order which is usually used in multitree plotting.
        """

        # pop fill from color dict if using color
        tstyle = deepcopy(self.style.tip_labels_style)
        if isinstance(self.style.tip_labels_colors, np.ndarray):
            tstyle.pop("fill")

        # circular layout ()
        if self.style.layout == 'c':

            # expand colors to an array so it can be masked if needed
            if isinstance(self.style.tip_labels, (np.ndarray, list, tuple)):
                colors = np.array(self.style.tip_labels_colors)
            else:
                colors = np.array([self.style.tip_labels] * self.ttree.ntips)

            # get tip coords at radius edge or at tips
            if self.style.tip_labels_align:
                tipcoords = self.coords.circ.get_tip_end_coords()
            else:
                tipcoords = self.coords.verts[:self.ttree.ntips]

            # adjust for baseline (origin) shift
            tipcoords[:, 0] += self.style.xbaseline
            tipcoords[:, 1] += self.style.ybaseline

            # get tip angles using radians from origin
            angles = self.coords.circ.get_tip_end_angles() - 0.05

            # whether to flip tip labels
            mask = (
                (abs(self.coords.circ.tip_radians) > np.pi / 2.) & 
                (abs(self.coords.circ.tip_radians) < 3 * np.pi / 2.)
            )

            # print in tips in two-parts for orientations
            tstyle["text-anchor"] = "end"
            if "-" in tstyle["-toyplot-anchor-shift"]:
                tstyle["-toyplot-anchor-shift"] = tstyle["-toyplot-anchor-shift"][1:]
            else:
                tstyle["-toyplot-anchor-shift"] = "-" + tstyle["-toyplot-anchor-shift"]
            self.axes.text(
                tipcoords[:, 0][mask],
                tipcoords[:, 1][mask],                
                np.array(self.tip_labels)[mask],
                angle=angles[mask] + 180,
                style=tstyle, 
                color=(
                    None if not isinstance(self.style.tip_labels_colors, np.ndarray)
                    else self.style.tip_labels_colors[mask]
                    ),
            )

            mask = np.invert(mask)
            tstyle["text-anchor"] = "start"
            if "-" in tstyle["-toyplot-anchor-shift"]:
                tstyle["-toyplot-anchor-shift"] = tstyle["-toyplot-anchor-shift"][1:]
            else:
                tstyle["-toyplot-anchor-shift"] = "-" + tstyle["-toyplot-anchor-shift"]
            self.axes.text(
                tipcoords[:, 0][mask],
                tipcoords[:, 1][mask],                
                np.array(self.tip_labels)[mask],
                angle=angles[mask],
                style=tstyle,
                color=(
                    None if not isinstance(self.style.tip_labels_colors, np.ndarray)
                    else self.style.tip_labels_colors[mask]
                    ),
            )            

        # unrooted orientations
        elif self.style.layout == "x":
            raise NotImplementedError("unrooted layout coming soon.")
        #     angles = [
        #         self.get_angle(self.ttree.treenode.get_leaves_by_name(i)[0])
        #         for i in self.ttree.get_tip_labels()
        #     ]
        #     self.axes.text(
        #         self._unrooted_coords[:self.ttree.ntips, 0],
        #         self._unrooted_coords[:self.ttree.ntips, 1],                
        #         self.tip_labels,
        #         angle=angles,
        #         style=tstyle,
        #         color=self.style.tip_labels_colors,
        #     )

        # re-orient for rooted orientations
        else:
            # get tip-coords and replace if using fixed_order
            xpos = self.ttree.get_tip_coordinates('x')
            ypos = self.ttree.get_tip_coordinates('y')

            # order tips for direction            
            if self.style.layout in ("u", "d"):
                if self.ttree._fixed_order:
                    xpos = np.arange(self.ttree.ntips) + self.style.xbaseline
                    ypos = ypos[self.ttree._fixed_idx]
                if self.style.tip_labels_align:
                    ypos = np.zeros(self.ttree.ntips)

                # up-facing style overrides
                if self.style.layout == 'u':
                    tstyle["text-anchor"] = "end"
                    if tstyle["-toyplot-anchor-shift"][0] != "-":
                        tstyle["-toyplot-anchor-shift"] = (
                            "-" + tstyle["-toyplot-anchor-shift"])

            if self.style.layout in ("r", "l"):
                if self.ttree._fixed_order:
                    xpos = xpos[self.ttree._fixed_idx]
                    ypos = np.arange(self.ttree.ntips) + self.style.ybaseline
                if self.style.tip_labels_align:
                    xpos = np.zeros(self.ttree.ntips)

                # left-facing style overrides 
                if self.style.layout == "l":
                    tstyle["text-anchor"] = "end"
                    if tstyle["-toyplot-anchor-shift"][0] != "-":
                        tstyle["-toyplot-anchor-shift"] = (
                            "-" + tstyle["-toyplot-anchor-shift"])

            # add tip names to coordinates calculated above
            self.axes.text(
                xpos, 
                ypos,
                self.tip_labels,
                angle=(0 if self.style.layout in ("r", "l") else -90),
                style=tstyle,
                color=self.style.tip_labels_colors,
            )

        # get stroke-width for aligned tip-label lines (optional)
        # copy stroke-width from the edge_style unless user set it
        if not self.style.edge_align_style.get("stroke-width"):
            self.style.edge_align_style["stroke-width"] = (
                self.style.edge_style["stroke-width"])


    def add_tip_lines_to_axes(self):
        "add lines to connect tips to zero axis for tip_labels_align=True"

        # get tip-coords and align-coords from verts
        xpos, ypos, aedges, averts = self.get_tip_label_coords() 
        if self.style.tip_labels_align:
            self.axes.graph(
                aedges,
                vcoordinates=averts,
                estyle=self.style.edge_align_style, 
                vlshow=False,
                vsize=0,
            )


    def fit_tip_labels(self):
        """
        Modifies display range to ensure tip labels fit. This is a bit hackish
        still. The problem is that the 'extents' range of the rendered text
        is totally correct. So we add a little buffer here. Should add for 
        user to be able to modify this if needed. If not using edge lengths
        then need to use unit length for treeheight.
        """
        # bail on unrooted for now; TODO
        if self.style.layout == "c":
            return

        # longest name
        lname = max([len(i) for i in self.tip_labels])

        # get ratio of names to tree in plot
        ratio = max(lname / 10, 0.15)

        # have tree figure make up 85% of plot
        if self.style.use_edge_lengths:
            addon = self.ttree.treenode.height
        else:
            addon = self.ttree.treenode.get_farthest_leaf(True)[1] + 1
        addon *= ratio

        # modify display for layout
        if self.style.tip_labels:
            if self.style.layout == "r":
                self.axes.x.domain.max = addon / 2.
                if self.style.xbaseline:
                    self.axes.x.domain.max += self.style.xbaseline
            elif self.style.layout == "l":
                self.axes.x.domain.min = (-1 * addon) / 2.
                if self.style.xbaseline:
                    self.axes.x.domain.min -= self.style.xbaseline
            elif self.style.layout == "d":
                self.axes.y.domain.min = (-1 * addon) / 2
                if self.style.ybaseline:
                    self.axes.y.domain.min += self.style.ybaseline
            elif self.style.layout == "u":
                self.axes.y.domain.max = (addon) / 2
                if self.style.ybaseline:
                    self.axes.y.domain.max += self.style.ybaseline


    def assign_node_colors_and_style(self):
        """
        Resolve conflict of 'node_color' and 'node_style['fill'] args which are
        redundant. Default is node_style.fill unless user entered node_color.
        To enter multiple colors user must use node_color not style fill. 
        Either way, we build a list of colors to pass to Drawing.node_colors 
        which is then written to the marker as a fill CSS attribute.
        """
        # SET node_colors and POP node_style.fill
        colors = self.style.node_colors
        style = self.style.node_style
        if colors is None:
            if style["fill"] in (None, "none"):
                style.pop("fill")
            else:
                if isinstance(style["fill"], ITERABLE):
                    raise ToytreeError(
                        "Use node_color not node_style for multiple node colors")
                # check the color
                color = style["fill"]
                if isinstance(color, (np.ndarray, np.void, list, tuple)):
                    color = toyplot.color.to_css(color)
                self.node_colors = [color] * self.ttree.nnodes

        # otherwise parse node_color
        else:
            style.pop("fill")
            if isinstance(colors, str):
                # check the color
                color = colors
                if isinstance(color, (np.ndarray, np.void, list, tuple)):
                    color = toyplot.color.to_css(color)
                self.node_colors = [color] * self.ttree.nnodes

            elif isinstance(colors, ITERABLE):
                if len(colors) != len(self.node_colors):
                    raise ToytreeError("node_colors arg is the wrong length")
                for cidx in range(len(self.node_colors)):
                    color = colors[cidx]
                    if isinstance(color, (np.ndarray, np.void, list, tuple)):
                        color = toyplot.color.to_css(color)                   
                    self.node_colors[cidx] = color

        # use CSS none for stroke=None
        if self.style.node_style["stroke"] is None:
            self.style.node_style.stroke = "none"

        # apply node markers
        markers = self.style.node_markers
        if markers is None:
            self.node_markers = ["o"] * self.ttree.nnodes
        else:
            if isinstance(markers, str):
                self.node_markers = [markers] * self.ttree.nnodes
            elif isinstance(markers, ITERABLE):
                for cidx in range(len(self.node_markers)):
                    self.node_markers[cidx] = markers[cidx]


    def assign_node_labels_and_sizes(self):
        "assign features of nodes to be plotted based on user kwargs"

        # shorthand
        nvals = self.ttree.get_node_values()

        # False == Hide nodes and labels unless user entered size 
        if self.style.node_labels is False:
            self.node_labels = ["" for i in nvals]           
            if self.style.node_sizes is not None:
                if isinstance(self.style.node_sizes, ITERABLE):
                    assert len(self.node_sizes) == len(self.style.node_sizes)
                    self.node_sizes = self.style.node_sizes

                elif isinstance(self.style.node_sizes, (int, str)):
                    self.node_sizes = (
                        [int(self.style.node_sizes)] * len(nvals)
                    )
                self.node_labels = [" " if i else "" for i in self.node_sizes]


        # True == Show nodes, label=idx, and show hover
        elif self.style.node_labels is True:
            # turn on node hover even if user did not set it explicit
            self.style.node_hover = True

            # get idx labels
            self.node_labels = self.ttree.get_node_values('idx', 1, 1)

            # use default node size as a list if not provided
            if not self.style.node_sizes:
                self.node_sizes = [18] * len(nvals)
            else:
                assert isinstance(self.style.node_sizes, (int, str))
                self.node_sizes = (
                    [int(self.style.node_sizes)] * len(nvals)
                )

        # User entered lists or other for node labels or sizes; check lengths.
        else:
            # make node labels into a list of values 
            if isinstance(self.style.node_labels, ITERABLE):
                assert len(self.style.node_labels) == len(nvals)
                self.node_labels = self.style.node_labels

            # check if user entered a feature else use entered val
            elif isinstance(self.style.node_labels, str):
                self.node_labels = [self.style.node_labels] * len(nvals)
                if self.style.node_labels in self.ttree.features:
                    self.node_labels = self.ttree.get_node_values(
                        self.style.node_labels, 1, 0)

            # default to idx at internals if nothing else
            else:
                self.node_labels = self.ttree.get_node_values("idx", 1, 0)

            # make node sizes as a list; set to zero if node label is ""
            if isinstance(self.style.node_sizes, ITERABLE):
                assert len(self.style.node_sizes) == len(nvals)
                self.node_sizes = self.style.node_sizes
            elif isinstance(self.style.node_sizes, (str, int, float)):
                self.node_sizes = [int(self.style.node_sizes)] * len(nvals)
            else:
                self.node_sizes = [18] * len(nvals)

            # override node sizes to hide based on node labels
            for nidx, node in enumerate(self.node_labels):
                if self.node_labels[nidx] == "":
                    self.node_sizes[nidx] = 0

        # ensure string type
        self.node_labels = [str(i) for i in self.node_labels]


    def assign_tip_labels_and_colors(self):
        "assign tip labels based on user provided kwargs"

        # COLOR (make a list from a single value)
        if isinstance(self.style.tip_labels_colors, str):
            self.style.tip_labels_colors = np.array(
                [self.style.tip_labels_colors] * self.ttree.ntips)

        # reorder array for fixed if needed
        elif isinstance(self.style.tip_labels_colors, ITERABLE):
            if self.ttree._fixed_order:
                cols = np.array(self.style.tip_labels_colors)
                orde = cols[self.ttree._fixed_idx]
                self.style.tip_labels_colors = list(orde)
            self.style.tip_labels_colors = np.array(self.style.tip_labels_colors)

        # LABELS
        # False == hide tip labels
        if self.style.tip_labels is False:
            self.style.tip_labels_style["-toyplot-anchor-shift"] = "0px"
            self.tip_labels = ["" for i in self.ttree.get_tip_labels()]

        # LABELS
        # user entered something...
        else:
            # if user did not change label-offset then shift it here
            if not self.style.tip_labels_style["-toyplot-anchor-shift"]:
                self.style.tip_labels_style["-toyplot-anchor-shift"] = "15px"

            # if user entered list in get_tip_labels order reverse it for plot
            if isinstance(self.style.tip_labels, ITERABLE):
                self.tip_labels = self.style.tip_labels

            # True assigns tip labels from tree
            else:
                if self.ttree._fixed_order:
                    self.tip_labels = self.ttree._fixed_order
                else:
                    self.tip_labels = self.ttree.get_tip_labels()


    def assign_edge_colors_and_widths(self):
        """
        Resolve conflict of 'node_color' and 'node_style['fill'] args which are
        redundant. Default is node_style.fill unless user entered node_color.
        To enter multiple colors user must use node_color not style fill. 
        Either way, we build a list of colors to pass to Drawing.node_colors 
        which is then written to the marker as a fill CSS attribute.
        """
        # node_color overrides fill. Tricky to catch cuz it can be many types.

        # SET edge_widths and POP edge_style.stroke-width
        if self.style.edge_widths is None:
            if not self.style.edge_style["stroke-width"]:
                self.style.edge_style.pop("stroke-width")
                self.style.edge_style.pop("stroke")
                self.edge_widths = [None] * self.nedges
            else:
                if isinstance(self.style.edge_style["stroke-width"], ITERABLE):
                    raise ToytreeError(
                        "Use edge_widths not edge_style for multiple edge widths")
                # check the color
                width = self.style.edge_style["stroke-width"]
                self.style.edge_style.pop("stroke-width")
                self.edge_widths = [width] * self.nedges
        else:
            self.style.edge_style.pop("stroke-width")            
            if isinstance(self.style.edge_widths, (str, int)):

                # Special case of setting Ne for plotting
                if self.style.edge_widths == "Ne":
                    if not all([i for i in self.ttree.get_edge_values("Ne")]):
                        self.edge_widths = np.repeat(2, self.nedges)
                    else:
                        self.edge_widths = (self.ttree.get_edge_values(
                            feature="Ne", normalize=True))

                # normal value entry
                else:
                    self.edge_widths = (
                        [int(self.style.edge_widths)] * self.nedges)

            elif isinstance(self.style.edge_widths, ITERABLE):
                if len(self.style.edge_widths) != self.nedges:
                    raise ToytreeError("edge_widths arg is the wrong length")
                for cidx in range(self.nedges):
                    self.edge_widths[cidx] = self.style.edge_widths[cidx]

        # SET edge_colors and POP edge_style.stroke
        if self.style.edge_colors is None:
            if self.style.edge_style["stroke"] is None:
                self.style.edge_style.pop("stroke")
                self.edge_colors = [None] * self.nedges
            else:
                if isinstance(self.style.edge_style["stroke"], ITERABLE):
                    raise ToytreeError(
                        "Use edge_colors not edge_style for multiple edge colors")
                # check the color
                color = self.style.edge_style["stroke"]
                if isinstance(color, (np.ndarray, np.void, list, tuple)):
                    color = toyplot.color.to_css(color)
                self.style.edge_style.pop("stroke")                    
                self.edge_colors = [color] * self.nedges

        # otherwise parse node_color
        else:
            self.style.edge_style.pop("stroke")                                
            if isinstance(self.style.edge_colors, (str, int)):
                # check the color
                color = self.style.edge_colors
                if isinstance(color, (np.ndarray, np.void, list, tuple)):
                    color = toyplot.color.to_css(color)
                self.edge_colors = [color] * self.nedges

            elif isinstance(self.style.edge_colors, ITERABLE):
                if len(self.style.edge_colors) != self.nedges:
                    raise ToytreeError("edge_colors arg is the wrong length")
                for cidx in range(self.nedges):
                    self.edge_colors[cidx] = self.style.edge_colors[cidx]

        # do not allow empty edge_colors or widths
        self.edge_colors = [i if i else "#262626" for i in self.edge_colors]
        self.edge_widths = [i if i else 2 for i in self.edge_widths]

    # -----------------------------------------------------------------
    # Tree / Graph plotting
    # -----------------------------------------------------------------
    def add_tree_to_axes(self):
        """
        Edge types affect whether nodes look like V or U 
        """

        # if edge color or widths then override style stroke and stroke-width
        if self.style.edge_type in ('c'):
            self.axes.graph(
                self.coords.edges,
                vcoordinates=self.coords.verts,
                layout=toyplot.layout.IgnoreVertices(),
                vlshow=False,
                vsize=0,
                estyle=self.style.edge_style,
                ewidth=self.edge_widths,
                ecolor=self.edge_colors,
            )

        # default edge type is 'p' splitting
        else:
            self.expand_edges_to_lines("edge_colors")
            self.expand_edges_to_lines("edge_widths")
            # print(self.coords.lines.round(2))
            # print(self.coords.coords.round(2))
            self.axes.graph(
                self.coords.lines,
                vcoordinates=self.coords.coords,
                layout=toyplot.layout.IgnoreVertices(),
                vlshow=False,
                vsize=0.,
                estyle=self.style.edge_style, 
                ewidth=self.edge_widths,
                ecolor=self.edge_colors,
            )


    def expand_edges_to_lines(self, attr):
        """
        used for 'p' edge_type to expand edge styles to connecting edges.
        """
        # set default values
        if attr == "edge_colors":
            arr = ["#262626"] * self.coords.lines.shape[0]
        else:
            arr = [2] * self.coords.lines.shape[0]

        # build dict from edges
        cdict = {
            self.coords.edges[i, 1]: getattr(self, attr)[i]
            for i in range(len(self.coords.edges))
        }

        # build new values from lines
        for idx in range(len(arr)):
            edge = self.coords.lines[idx]

            # this edge goes into x
            into = edge[1]

            # colors going into x in edges is
            val = cdict.get(into)

            if val:
                # apply this val to every line into x
                lidx = np.where(self.coords.lines[:, 1] == into)
                arr[lidx[0][0]] = val

                # and into y
                y = edge[0]
                lidx = np.where(self.coords.lines[:, 1] == y)
                arr[lidx[0][0]] = val

        # update the value list        
        setattr(self, attr, arr)  # edge_colors = arr

    # -----------------------------------------------------------------
    # Node and Node Labels 
    # -----------------------------------------------------------------   
    def add_nodes_to_axes(self):
        """
        Creates a new marker for every node from idx indexes and lists of 
        node_values, node_colors, node_sizes, node_style, node_labels_style.
        Pulls from node_color and adds to a copy of the style dict for each 
        node to create marker.

        Node_colors has priority to overwrite node_style['fill']
        """
        # bail out if not any visible nodes (e.g., none w/ size>0)
        if all([i == "" for i in self.node_labels]):
            return

        # build markers for each node.
        marks = []
        for nidx in self.ttree.get_node_values('idx', 1, 1):

            # select node value from deconstructed lists
            nlabel = self.node_labels[nidx]
            nsize = self.node_sizes[nidx]
            nmarker = self.node_markers[nidx]

            # get styledict copies
            nstyle = deepcopy(self.style.node_style)
            nlstyle = deepcopy(self.style.node_labels_style)

            # and mod style dict copies from deconstructed lists
            nstyle["fill"] = self.node_colors[nidx]

            # create mark if text or node
            if (nlabel or nsize):
                mark = toyplot.marker.create(
                    shape=nmarker, 
                    label=str(nlabel),
                    size=nsize,
                    mstyle=nstyle,
                    lstyle=nlstyle,
                )
            else:
                mark = ""

            # store the nodes/marks
            marks.append(mark)

        # node_hover == True to show all features interactive
        if self.style.node_hover is True:
            title = self.get_hover()

        elif isinstance(self.style.node_hover, list):
            # todo: return advice if improperly formatted
            title = self.style.node_hover

        # if hover is false then no hover
        else:
            title = None

        # layout changes for unrooted trees
        if self.style.edge_type == 'u':
            xcoords = self._unrooted_coords[:, 0]
            ycoords = self._unrooted_coords[:, 1]
        else:
            xcoords = self.coords.verts[:, 0]
            ycoords = self.coords.verts[:, 1]

        # add nodes
        self.axes.scatterplot(
            xcoords, 
            ycoords, 
            marker=marks,
            title=title,
        )

    # -----------------------------------------------------------------
    # Axes styling / scale bar / padding
    # -----------------------------------------------------------------        
    def add_axes_style(self):

        # style axes with padding and show axes
        self.axes.padding = self.style.padding

        if not self._external_axis:
            self.axes.show = True
            if not self.style.scalebar:
                self.axes.show = False

        # scalebar        
        if self.style.scalebar and self.style.layout == "r":
            nticks = max((3, np.floor(self.style.width / 100).astype(int)))
            self.axes.y.show = False
            self.axes.x.show = True
            self.axes.x.ticks.show = True

            # generate locations
            if self.style.use_edge_lengths:
                locs = np.linspace(0, self.ttree.treenode.height, nticks) * -1
            else:
                top = self.ttree.treenode.get_farthest_leaf(True)[1] + 1
                locs = np.linspace(0, top, nticks) * -1

            # auto-formatter for axes ticks labels
            zer = abs(min(0, Decimal(locs[1]).adjusted()))
            fmt = "{:." + str(zer) + "f}"
            self.axes.x.ticks.locator = toyplot.locator.Explicit(
                locations=locs,
                labels=[fmt.format(i) for i in np.abs(locs)],
                )

        elif self.style.scalebar and self.style.layout == "d":
            nticks = max((3, np.floor(self.style.height / 100).astype(int)))
            self.axes.x.show = False
            self.axes.y.show = True
            self.axes.y.ticks.show = True            

            # generate locations
            locs = np.linspace(0, self.ttree.treenode.height, nticks)

            # auto-formatter for axes ticks labels
            zer = abs(min(0, Decimal(locs[1]).adjusted()))
            fmt = "{:." + str(zer) + "f}"
            self.axes.y.ticks.locator = toyplot.locator.Explicit(
                locations=locs,
                labels=[fmt.format(i) for i in np.abs(locs)],
                )

    # ------------------------------------------------------------------
    # Other helper functions
    # ------------------------------------------------------------------
    def get_tip_label_coords(self):
        """
        Get starting position of tip labels text based on locations of the 
        leaf nodes on the tree and style offset and align options. Node
        positions are found using the .verts attribute of coords and is 
        already oriented for the tree face direction. 
        """
        # number of tips
        ns = self.ttree.ntips

        # x-coordinate of tips assuming down-face
        tip_xpos = self.coords.verts[:ns, 0]
        tip_ypos = self.coords.verts[:ns, 1]
        align_edges = None
        align_verts = None

        # cirulcar name layout
        if self.style.layout == "c":
            if self.style.tip_labels_align:

                # get tip coordinates as x,y
                endcoords = self.coords.circ.get_tip_end_coords()
                tipcoords = self.coords.verts[:self.ttree.ntips]
                endcoords[:, 0] += self.ttree.style.xbaseline
                endcoords[:, 1] += self.ttree.style.ybaseline                

                # the array of which tips go to which tip terminal
                align_edges = np.array(
                    [(i, i + len(endcoords)) for i in range(len(endcoords))]
                )

                # the vertex locations of tips and tip terminals.
                align_verts = np.concatenate([tipcoords, endcoords])
                return None, None, align_edges, align_verts


        # handle layout orientations
        if self.style.layout in ('u', 'd'):
            # align tips at zero
            if self.style.tip_labels_align:
                tip_yend = np.zeros(ns)
                align_edges = np.array([
                    (i + len(tip_ypos), i) for i in range(len(tip_ypos))
                ])
                align_verts = np.array(
                    list(zip(tip_xpos, tip_ypos)) + \
                    list(zip(tip_xpos, tip_yend))
                )
                tip_ypos = tip_yend


        # other layouts: ('r', 'l')
        else:
            # tip labels align finds the zero axis for orientation...
            if self.style.tip_labels_align:
                tip_xend = np.zeros(ns)
                align_edges = np.array([
                    (i + len(tip_xpos), i) for i in range(len(tip_xpos))
                ])
                align_verts = np.array(
                    list(zip(tip_xpos, tip_ypos)) + \
                    list(zip(tip_xend, tip_ypos))
                )
                tip_xpos = tip_xend

        return tip_xpos, tip_ypos, align_edges, align_verts


    def get_hover(self, ordered_features=["idx", "name", "dist", "support"]):
        # build full features titles
        lfeatures = list(set(self.ttree.features) - set(ordered_features))       
        ordered_features += lfeatures

        # build list of hoverstrings in order of idxs
        title = [" "] * self.ttree.nnodes
        for nidx in self.ttree.get_node_values('idx', True, True):
            feats = []
            for feature in ordered_features:
                val = self.ttree.get_node_values(feature, True, True)[nidx]
                if isinstance(val, float):
                    feats.append("{}: {:.4f}".format(feature, val))
                else:
                    feats.append("{}: {}".format(feature, val))
            title[nidx] = feats

        # concatenate into strings for each sample
        title = ["\n".join(i) for i in title]

        # order of hover can depend on the orientation...
        title = title[::-1]
        return title


    def get_dims_from_tree_size(self):
        "Calculate reasonable canvas height and width for tree given N tips" 

        if self.style.layout == "c":
            if not self.style.height:
                self.style.height = 300
            if not self.style.width:
                self.style.width = 300
            return 

        lname = max([len(i) for i in self.tip_labels])     
        if self.style.layout in ("r", "l"):
            # height fit by tree size
            if not self.style.height:
                self.style.height = max(275, min(1000, 18 * self.ttree.ntips))
            # width fit by name size
            if not self.style.width:
                self.style.width = max(250, min(500, 250 + 5 * lname))
        else:
            # height fit by name size
            if not self.style.height:
                self.style.height = max(250, min(500, 250 + 5 * lname))
            # width fit by tree size
            if not self.style.width:
                self.style.width = max(350, min(1000, 18 * self.ttree.ntips))


    def get_canvas_and_axes(self, axes):
        if axes: 
            self.canvas = None
            self.axes = axes
            self._external_axis = True
        else:
            self.canvas = toyplot.Canvas(
                height=self.style.height,
                width=self.style.width,
            )
            self.axes = self.canvas.cartesian(
                padding=self.style.padding
            )

        # return nothing if tree is empty
        if not self.ttree.treenode.children:
            raise ToytreeError("Tree is empty")


    def add_admixture_edges(self):
        if self.style.admixture_edges is not None:
            AdmixEdges(
                self._tree, 
                self.axes, 
                self.style.admixture_edges,
                layout=self.style.layout,
                edge_type=self.style.edge_type,
            )


    # not being used, not right
    def get_angle2(self, node):
        n1 = node.idx
        n2 = node.up.idx
        c1 = self._unrooted_coords[n1]
        c2 = self._unrooted_coords[n2]
        dy = c1[1] - c2[1]
        radians = np.arccos(dy / node.dist)
        if radians > 0:
            degrees = np.rad2deg(radians)
        else:
            degrees = 360 + np.rad2deg(radians)
        return degrees


    def get_angle(self, node):       
        # get coordinates
        n1 = node.idx
        n2 = node.up.idx
        c1 = self._unrooted_coords[n1]
        c2 = self._unrooted_coords[n2]
        dx = c1[0] - c2[0]
        dy = c1[1] - c2[1]

        # opposite over adjancent
        radprop = np.arctan(dy / dx)
        degprop = radprop / (np.pi / 2.)
        deg = 90 * degprop

        # get angle 
        if dx < 0:
            if dy < 0:
                deg += 180
            else:
                deg += 180
        else:
            if dy < 0:
                deg += 360
            else:
                deg += 0
        return deg



#!/usr/bin/env python

"MultiTree objects"

from __future__ import print_function, absolute_import
from builtins import range, str

from copy import deepcopy
from hashlib import md5
from collections import defaultdict

import toyplot
import toyplot.config
import numpy as np

# used in Consensus
from .TreeNode import TreeNode
from .Toytree import ToyTree
from .TreeParser import TreeParser
from .TreeStyle import TreeStyle, STYLES
from .MultiDrawing import TreeGrid, CloudTree
from .utils import bpp2newick



class MultiTree(object):
    """
    Toytree MultiTree object for representing multiple trees. 

    Parameters:
    -----------
    newick: (str)
        string, filepath, or URL for a newick or nexus formatted list of trees
    tree_format: (int)
        ete format for newick tree structure. Default is 0. 
    fixed_order: (bool, list, None)    
        ...

    Attributes:
    -----------
    treelist: list
        A list of toytree objects from the parsed newick file. 

    Functions():
    ------------
    get_consensus_tree: 
        Returns a ToyTree object with support values on nodes.
    draw_cloud_tree:
        Draws a plot with overlapping fixed_order trees.
    draw_grid_tree:
        Draws a plot with n x m trees in a grid.
    """
    def __init__(self, newick, tree_format=0):  # , fixed_order=False):

        # setting attributes
        self.style = TreeStyle('m')
        self._i = 0

        # parse the newick object into a list of Toytrees
        self.treelist = []
        if isinstance(newick, str):
            self.treelist = [
                ToyTree(i) for i in 
                TreeParser(newick, tree_format, multitree=True).treenodes
            ]

        # iterables (list, tuple, ndarray, Series)
        else:
            # convert to list
            if newick is not None:
                newick = list(newick)
            # load list whether it is newicks, toytrees or treenodes
            if isinstance(newick[0], str):
                self.treelist = [
                    ToyTree(i) for i in 
                    TreeParser(newick, tree_format, multitree=True).treenodes
                ]
            elif isinstance(newick[0], ToyTree):
                self.treelist = newick
            elif isinstance(newick[0], TreeNode):
                self.treelist = [ToyTree(i) for i in newick]

        # set tip plot order for treelist to the first tree order
        # order trees in treelist to plot in shared order...
        # self._fixed_order = fixed_order   # (<list>, True, False, or None)
        # self._user_order = None
        # self._cons_order = None
        # self._set_tip_order()
        # self._parse_treelist()

    # attributes of multitrees
    def __len__(self):  
        return len(self.treelist)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            result = self.treelist[self._i]
        except IndexError:
            self._i = 0
            raise StopIteration
        self._i += 1
        return result

    @property
    def ntips(self):
        return self.treelist[0].ntips


    @property
    def ntrees(self):
        return len(self.treelist)


    @property
    def all_tips_shared(self):
        #if names are the same in all the trees...
        alltips_shared = all([
            set(self.treelist[0].get_tip_labels()) == set(i.get_tip_labels()) 
            for i in self.treelist
        ])
        if alltips_shared:
            return True
        return False


    def copy(self):
        return deepcopy(self)


    def write(self, handle=None, format=0):
        if not handle:
            handle = "out.tre"
        with open(handle, 'w') as outtre:
            for tre in self.treelist:
                outtre.write(tre.newick + "\n")


    def reset_tree_styles(self):
        """
        Sets the .style toytree drawing styles to default for all ToyTrees
        in a MultiTree .treelist. 
        """
        for tre in self.treelist:
            tre.style = TreeStyle('n')


    # -------------------------------------------------------------------
    # Tree List Statistics or Calculations
    # -------------------------------------------------------------------
    # def get_tip_labels(self):
    #     """
    #     Returns the tip names in tree plot order for the *list of tree*, 
    #     starting from the zero axis. If all trees in the treelist do not 
    #     share the same set of tips then this will return an error message. 

    #     If fixed_order is a user entered list then names are returned in that
    #     order. If fixed_order was True then the consensus tree order is 
    #     returned. If fixed_order was None or False then the order of the first
    #     ToyTree in .treelist is returned. 
    #     """
    #     if not self.all_tips_shared:
    #         raise Exception(
    #             "All trees in treelist do not share the same set of tips")
    #     return self.treelist[0].get_tip_labels()


    def get_consensus_tree(self, cutoff=0.0, best_tree=None):
        """
        Returns an extended majority rule consensus tree as a Toytree object.
        Node labels include 'support' values showing the occurrence of clades 
        in the consensus tree across trees in the input treelist. 
        Clades with support below 'cutoff' are collapsed into polytomies.
        If you enter an optional 'best_tree' then support values from
        the treelist calculated for clades in this tree, and the best_tree is
        returned with support values added to nodes. 

        Params
        ------
        cutoff (float; default=0.0): 
            Cutoff below which clades are collapsed in the majority rule 
            consensus tree. This is a proportion (e.g., 0.5 means 50%). 
        best_tree (Toytree or newick string; optional):
            A tree that support values should be calculated for and added to. 
            For example, you want to calculate how often clades in your best 
            ML tree are supported in 100 bootstrap trees. 
        """
        if best_tree is not None:
            if not isinstance(best_tree, ToyTree):
                best_tree = ToyTree(best_tree)
        cons = ConsensusTree(self.treelist, best_tree=best_tree, cutoff=cutoff)
        cons.update()
        return cons.ttree

    # -------------------------------------------------------------------
    # Tree List Plotting
    # -------------------------------------------------------------------
    def draw_tree_grid(
        self, 
        axes=None,
        nrows=None, 
        ncols=None, 
        start=0, 
        fixed_order=False, 
        shared_axis=False, 
        **kwargs):
        """        
        Draw a slice of x*y trees into a x,y grid non-overlapping. 

        Parameters:
        -----------
        x (int):
            Number of grid cells in x dimension. Default=automatically set.
        y (int):
            Number of grid cells in y dimension. Default=automatically set.
        start (int):
            Starting index of tree slice from .treelist.
        kwargs (dict):
            Toytree .draw() arguments as a dictionary. 
        """
        # return nothing if tree is empty
        if not self.treelist:
            print("Treelist is empty")
            return None, None

        # make a copy of the treelist so we don't modify the original
        if not fixed_order:
            treelist = self.copy().treelist
        else:
            if fixed_order is True:
                fixed_order = self.treelist[0].get_tip_labels()
            treelist = [
                ToyTree(i, fixed_order=fixed_order) 
                for i in self.copy().treelist
            ]

        # apply kwargs styles to the individual tree styles
        for tree in treelist:
            if kwargs.get("ts"):
                tree.style = TreeStyle(kwargs.get("ts"))
            if kwargs.get("tree_style"):
                tree.style = TreeStyle(kwargs.get("tree_style"))
            tree.style.update(kwargs)

        # get reasonable values for x,y given treelist length
        if not (ncols or nrows):
            nrows = 1
            if self.ntrees < 6:
                ncols = self.ntrees
            else:
                ncols = 5           

        # one or the other
        elif not (ncols and nrows):
            if ncols:
                if ncols == 1:
                    if self.ntrees <= 5:
                        nrows = self.ntrees
                    else:
                        nrows = 2
                else:
                    if self.ntrees <= 10:
                        nrows = 2
                    else:
                        nrows = 3

            if nrows:
                if nrows == 1:
                    if self.ntrees <= 5:
                        ncols = self.ntrees 
                    else:
                        ncols = 5
                else:
                    if self.ntrees <= 10:
                        ncols = 5
                    else:
                        ncols = 3
        else:
            pass

        # Return TereGrid object for debugging
        draw = TreeGrid(treelist)
        if kwargs.get("debug"):
            return draw

        # Call update to draw plot. Kwargs still here for width, height, axes
        canvas, axes = draw.update(
            axes, nrows, ncols, start, shared_axis, **kwargs)
        return canvas, axes



    def draw_cloud_tree(
        self, 
        axes=None, 
        html=False,
        fixed_order=True,
        **kwargs):
        """
        Draw a series of trees overlapping each other in coordinate space.
        The order of tip_labels is fixed in cloud trees so that trees with 
        discordant relationships can be seen in conflict. To change the tip
        order use the 'fixed_order' argument in toytree.mtree() when creating
        the MultiTree object.

        Parameters:
            axes (toyplot.Cartesian): toyplot Cartesian axes object.
            html (bool): whether to return the drawing as html (default=PNG).
            **kwargs (dict): styling options should be input as a dictionary.
        """
        # return nothing if tree is empty
        if not self.treelist:
            print("Treelist is empty")
            return None, None

        # return nothing if tree is empty
        if not self.all_tips_shared:
            print("All trees in treelist do not share the same tips")
            return None, None            

        # make a copy of the treelist so we don't modify the original
        if not fixed_order:
            raise Exception(
                "fixed_order must be either True or a list with the tip order")

        # set fixed order on a copy of the tree list
        if isinstance(fixed_order, (list, tuple)):
            fixed_order = fixed_order
        elif fixed_order is True:
            fixed_order = self.treelist[0].get_tip_labels()
        else:
            raise Exception(
                "fixed_order argument must be True or a list with the tip order")
        treelist = [
            ToyTree(i, fixed_order=fixed_order) for i in self.copy().treelist
        ]  

        # give advice if user tries to enter tip_labels
        if kwargs.get("tip_labels"):
            if not isinstance(kwargs.get("tip_labels"), dict):
                print(TIP_LABELS_ADVICE)
                kwargs.pop("tip_labels")

        # set autorender format to png so we don't bog down notebooks
        try:
            changed_autoformat = False
            if not html:
                toyplot.config.autoformat = "png"
                changed_autoformat = True

            # dict of global cloud tree style 
            mstyle = deepcopy(STYLES['m'])

            # if trees in treelist already have some then we don't quash...
            mstyle.update(
                {i: j for (i, j) in kwargs.items() if 
                (j is not None) & (i != "tip_labels")}
            )
            for tree in treelist:
                tree.style.update(mstyle)

            # Send a copy of MultiTree to init Drawing object.
            draw = CloudTree(treelist, **kwargs)

            # and create drawing
            if kwargs.get("debug"):
                return draw

            # allow user axes, and kwargs for width, height
            canvas, axes = draw.update(axes)
            return canvas, axes

        finally:
            if changed_autoformat:
                toyplot.config.autoformat = "html"


    # # allow ts as a shorthand for tree_style
    # if kwargs.get("ts"):
    #     tree_style = kwargs.get("ts")

    # # pass a copy of this tree so that any mods to .style are not saved
    # nself = deepcopy(self)
    # if tree_style:
    #     nself.style.update(TreeStyle(tree_style[0]))


class ConsensusTree:
    """
    An extended majority rule consensus function.
    Modelled on the similar function from scikit-bio tree module. If
    cutoff=0.5 then it is a normal majority rule consensus, while if
    cutoff=0.0 then subsequent non-conflicting clades are added to the tree.
    """
    def __init__(self, treelist, best_tree=None, cutoff=0.0):

        # parse args
        self.treelist = treelist
        self.best_tree = best_tree
        if self.best_tree is not None:
            self.best_tree = best_tree.copy().unroot()
            self.names = self.best_tree.get_tip_labels()
        else:
            self.names = self.treelist[0].get_tip_labels()
        self.cutoff = float(cutoff)

        # attrs to fill
        self.namedict = None
        self.treedict = {}
        self.clade_counts = None
        self.fclade_counts = None

        # results 
        self.ttree = None
        self.nodelist = None

        assert cutoff < 1, "cutoff should be a float proportion (e.g., 0.5)"

    def update(self):

        # hash a dict to remove duplicate trees
        self.hash_trees()

        # map onto best_tree of infer majrule consensus
        if self.best_tree is not None:
            self.map_onto_best_tree()

        else:
            # Find which clades occured with freq > cutoff. 
            # Fills namedict, clade_counts
            self.find_clades()

            # Filter out the < cutoff clades
            # Fills fclade_counts
            self.filter_clades()

            # Build consensus tree.
            # Fills .tree
            self.build_trees()  # fclade_counts, namedict)
            ## todo. make sure no singleton nodes were left behind ...


    def hash_trees(self):
        "hash ladderized tree topologies"       
        observed = {}
        for idx, tree in enumerate(self.treelist):
            nwk = tree.write(tree_format=9)
            hashed = md5(nwk.encode("utf-8")).hexdigest()
            if hashed not in observed:
                observed[hashed] = idx
                self.treedict[idx] = 1
            else:
                idx = observed[hashed]
                self.treedict[idx] += 1


    def map_onto_best_tree(self):
        "map clades from tree onto best_tree"

        # index names from the first tree
        ndict = {j: i for i, j in enumerate(self.names)}

        # dictionary of bits describing all clades in the best tree
        idict = {}
        bitdict = {}
        for node in self.best_tree.treenode.traverse("preorder"):

            # get byte string representing split
            bits = np.zeros(self.best_tree.ntips, dtype=np.bool_)
            for child in node.iter_leaf_names():
                bits[ndict[child]] = True
            bitstring = bits.tobytes()

            # record split (mirror image not relevant)
            bitdict[bitstring] = 0
            idict[bitstring] = node
        # print(bitdict)

        # count occurrence of clades in best_tree among other trees
        for tidx, ncopies in self.treedict.items():
            tre = self.treelist[tidx].unroot()
            # print(tidx)
            for node in tre.treenode.traverse("preorder"):
                bits = np.zeros(tre.ntips, dtype=np.bool_)
                for child in node.iter_leaf_names():
                    bits[ndict[child]] = True
                bitstring = bits.tobytes()
                # print(bits.astype(int))
                if bitstring in bitdict:
                    bitdict[bitstring] += ncopies
                else:
                    revstring = np.invert(bits).tobytes()
                    if revstring in bitdict:
                        bitdict[revstring] += ncopies
            # print("")

        # convert to frequencies
        for key, val in bitdict.items():
            # print(key, val, idict[key].name)
            idict[key].support = int(100 * val / float(len(self.treelist)))
        self.ttree = self.best_tree
        self.ttree._coords.update()


    def find_clades(self):
        "Count clade occurrences."
        # index names from the first tree
        ndict = {j: i for i, j in enumerate(self.names)}
        namedict = {i: j for i, j in enumerate(self.names)}

        # store counts
        clade_counts = {}
        for tidx, ncopies in self.treedict.items():

            # testing on unrooted trees is easiest but for some reason slow
            ttree = self.treelist[tidx].unroot()

            # traverse over tree
            for node in ttree.treenode.traverse('preorder'):
                bits = np.zeros(len(ttree), dtype=np.bool_)
                for child in node.iter_leaf_names():
                    bits[ndict[child]] = True

                # get bit string and its reverse
                bitstring = bits.tobytes()
                revstring = np.invert(bits).tobytes()

                # add to clades first time, then check for inverse next hits
                if bitstring in clade_counts:
                    clade_counts[bitstring] += ncopies
                else:
                    if revstring not in clade_counts:
                        clade_counts[bitstring] = ncopies
                    else:
                        clade_counts[revstring] += ncopies

        # convert to freq
        for key, val in clade_counts.items():
            clade_counts[key] = val / float(len(self.treelist))

        ## return in sorted order
        self.namedict = namedict
        self.clade_counts = sorted(
            clade_counts.items(),
            key=lambda x: x[1],
            reverse=True)


    def filter_clades(self):
        "Remove conflicting clades and those < cutoff to get majority rule"
        passed = []
        carrs = np.array([list(i[0]) for i in self.clade_counts], dtype=int)
        freqs = np.array([i[1] for i in self.clade_counts])

        for idx in range(carrs.shape[0]):
            conflict = False
            if freqs[idx] < self.cutoff:
                continue

            for pidx in passed:
                intersect = np.max(carrs[idx] + carrs[pidx]) > 1

                # is either one a subset of the other?
                subset_test0 = np.all(carrs[idx] - carrs[pidx] >= 0)
                subset_test1 = np.all(carrs[pidx] - carrs[idx] >= 0)
                if intersect:
                    if (not subset_test0) and (not subset_test1):
                        conflict = True

            if not conflict:
                passed.append(idx)

        rclades = []
        for idx in passed:
            rclades.append((carrs[idx], freqs[idx]))
        self.fclade_counts = rclades


    def build_trees(self):
        "Build an unrooted consensus tree from filtered clade counts."

        # storage
        nodes = {}
        idxarr = np.arange(len(self.fclade_counts[0][0]))
        queue = []

        # create dict of clade counts and set keys
        countdict = defaultdict(int)
        for clade, count in self.fclade_counts:
            mask = np.int_(list(clade)).astype(np.bool)
            ccx = idxarr[mask]
            queue.append((len(ccx), frozenset(ccx)))
            countdict[frozenset(ccx)] = count

        while queue:
            queue.sort()
            (clade_size, clade) = queue.pop(0)
            new_queue = []

            # search for ancestors of clade
            for (_, ancestor) in queue:
                if clade.issubset(ancestor):
                    # update ancestor such that, in the following example:
                    # ancestor == {1, 2, 3, 4}
                    # clade == {2, 3}
                    # new_ancestor == {1, {2, 3}, 4}
                    new_ancestor = (ancestor - clade) | frozenset([clade])
                    countdict[new_ancestor] = countdict.pop(ancestor)
                    ancestor = new_ancestor

                new_queue.append((len(ancestor), ancestor))

            # if the clade is a tip, then we have a name
            if clade_size == 1:
                name = list(clade)[0]
                name = self.namedict[name]
            else:
                name = None

            # the clade will not be in nodes if it is a tip
            children = [nodes.pop(c) for c in clade if c in nodes]
            node = TreeNode(name=name)
            for child in children:
                node.add_child(child)
            if not node.is_leaf():
                node.dist = int(round(100 * countdict[clade]))
                node.support = int(round(100 * countdict[clade]))
            else:
                node.dist = int(100)
                node.support = int(100)

            nodes[clade] = node
            queue = new_queue
        nodelist = list(nodes.values())
        tre = nodelist[0]

        ## return the tree and other trees if present
        self.ttree = ToyTree(tre.write(format=0))
        self.ttree._coords.update()
        self.nodelist = nodelist



TIP_LABELS_ADVICE = """
Warning: ignoring 'tip_labels' argument. 

The 'tip_labels' arg to draw_cloud_tree() should be a dictionary
mapping tip names from the contained ToyTrees to a new string value.
Example: {"a": "tip-A", "b": "tip-B", "c": tip-C"}

# get a MultiTree containing 10 trees with numbered tip names
trees = toytree.mtree([toytree.rtree.imbtree(5) for i in range(10)])

# draw a cloud tree using a set tip order and styled tip names
trees.draw_cloud_tree(
    fixed_order=['0', '1', '2', '3', '4'],
    tip_labels={i: "tip-{}".format(i) for i in trees.treelist[0].get_tip_labels()},
    )
"""



#!/usr/bin/env python

"""
Write Toytree or TreeNode to newick string
"""

import re
from .utils import NW_FORMAT

# globals
ILLEGAL_NEWICK_CHARS = r":;(),\[\]\t\n\r="
ITERABLE_TYPES = set([list, set, tuple, frozenset])

# formatters
FF = "%0.6g"
NF = "%s"
DF = FF  # ":" + FF


class NewickWriter(object):
    """
    Write TreeNode to a newick string.

    Parameters
    ----------
    tree: TreeNode
        Input TreeNode
    tree_format: (int)
        Format of newick string, see ete formats 0-10.
    format_root_node: (bool)
        Write node info (e.g., support, dist) for root node.   
    features: (None or list,set,tuple)
        Features to be written in NHX format in newick string.
    """
    def __init__(
        self, 
        treenode,
        tree_format=0, 
        format_root_node=False, 
        features=None,
        is_leaf_fn=None,
        dist_formatter=None,
        name_formatter=None,
        support_formatter=None,
        ):

        # store attrs
        self.treenode = treenode
        self.tf = tree_format
        self.format_root_node = format_root_node
        self.features = features
        self.fn = (is_leaf_fn if is_leaf_fn else is_leaf)
        self.newick = []

        # formatters
        self.dist_formatter = dist_formatter
        self.support_formatter = support_formatter
        self.name_formatter = name_formatter


    def write_newick(self):

        # returns (1, tipnode) or (0, tipnode); 1 means visited in postorder
        for post, node in self.treenode.iter_prepostorder(is_leaf_fn=self.fn):

            # if node visited in postorder then it has already been formatted
            if post:

                # close the node
                self.newick.append(")")

                # add node info and features
                if node.up or self.format_root_node:
                    self.newick.append(
                        format_node(
                            node, 
                            "internal", 
                            self.tf, 
                            self.dist_formatter,
                            self.support_formatter,
                            self.name_formatter,                            
                        )
                    )
                    self.newick.append(get_features(node, self.features))

            # first pass, node needs to be formatted...
            else:

                # if not root and not child0, put comma and wait for sister
                if (node != self.treenode) and (node != node.up.children[0]):
                    self.newick.append(",")

                # if leaf then write node name
                if node.is_leaf():

                    # sub out bad characters
                    node.name = re.sub(
                        "[" + ILLEGAL_NEWICK_CHARS + "]", 
                        "_", 
                        str(getattr(node, "name")),
                    )

                    # store formatted node
                    self.newick.append(
                        format_node(node, "leaf", self.tf, DF, DF, NF)
                    )
                    self.newick.append(get_features(node, self.features))

                # start new internal node
                else:
                    self.newick.append("(")

        self.newick.append(";")
        return ''.join(self.newick)



def format_node(
    node, 
    node_type, 
    format,
    dist_formatter=None,
    support_formatter=None,
    name_formatter=None):
    """
    Used when writing newick...
    """

    # use default formatters unless specified
    if dist_formatter is None: 
        dist_formatter = FF
    if support_formatter is None: 
        support_formatter = FF
    if name_formatter is None: 
        name_formatter = NF

    # apply formatting to leaves or nodes
    if node_type == "leaf":
        container1 = NW_FORMAT[format][0][0]  # name
        container2 = NW_FORMAT[format][1][0]  # dists
        converterFn1 = NW_FORMAT[format][0][1]
        converterFn2 = NW_FORMAT[format][1][1]
        flexible1 = NW_FORMAT[format][0][2]
    else:
        container1 = NW_FORMAT[format][2][0]  # support/name
        container2 = NW_FORMAT[format][3][0]  # dist
        converterFn1 = NW_FORMAT[format][2][1]
        converterFn2 = NW_FORMAT[format][3][1]
        flexible1 = NW_FORMAT[format][2][2]

    # Format node names
    FIRST_PART = ""
    if converterFn1 is not None:

        # convert node attribute to a string
        if converterFn1 == str:
            try:
                FIRST_PART = re.sub(
                    "[" + ILLEGAL_NEWICK_CHARS + "]",
                    "_", 
                    str(getattr(node, container1))
                )
                if not FIRST_PART and container1 == 'name' and not flexible1:
                    FIRST_PART = "NoName"
            except (AttributeError, TypeError):
                FIRST_PART = "?"

            # format as ...
            FIRST_PART = name_formatter % FIRST_PART

        # apply converter function to node attribute
        else:
            try:
                #FIRST_PART =  "%0.6f" %(converterFn2(getattr(node, container1)))
                FIRST_PART = support_formatter % (
                    converterFn2(getattr(node, container1))
                    )
            except (ValueError, TypeError):
                FIRST_PART = "?"

    # Format node info
    SECOND_PART = ""
    if converterFn2 is not None:

        # convert to string
        if converterFn2 == str:
            try:
                SECOND_PART = ":" + re.sub(
                    "[" + ILLEGAL_NEWICK_CHARS + "]",
                    "_",
                    str(getattr(node, container2))
                )
            except (ValueError, TypeError):
                SECOND_PART = ":?"
        else:
            try:
                SECOND_PART = ":%s" % (
                    dist_formatter % (converterFn2(getattr(node, container2))))
            except (ValueError, TypeError):
                SECOND_PART = ":?"

    return "%s%s" % (FIRST_PART, SECOND_PART)


def get_features(self, features=None):
    """ 
    Generates the extended newick string NHX with extra data about a node.
    """
    string = ""
    if features is None:
        features = []
    elif (features == []) or (features is True):
        features = self.features

    for pr in features:
        if hasattr(self, pr):
            raw = getattr(self, pr)
            
            # convert to a string
            if type(raw) in ITERABLE_TYPES:
                raw = '|'.join([str(i) for i in raw])
            elif type(raw) == dict:
                raw = '|'.join(
                    ["{}-{}".format(i, j) for (i, j) in raw.items()]
                )
                # map(lambda x,y: "%s-%s" %(x, y), six.iteritems(raw)))
            elif type(raw) == str:
                pass
            else:
                raw = str(raw)

            value = re.sub("[" + ILLEGAL_NEWICK_CHARS + "]", "_", raw)
            if string != "":
                string += ":"
            string += r"%s=%s" % (pr, str(value))
    if string != "":
        string = "[&&NHX:" + string + "]"

    return string




def is_leaf(node):
    return not bool(node.children)



#!/usr/bin/env python

from setuptools import setup, find_packages
import re

# get version from __init__.py
INITFILE = "toytree/__init__.py"
CUR_VERSION = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                    open(INITFILE, "r").read(),
                    re.M).group(1)

# run setup
setup(
    name="toytree",
    version=CUR_VERSION,
    url="https://github.com/eaton-lab/toytree",
    author="Deren Eaton",
    author_email="de2356@columbia.edu",
    description="minimalist tree plotting using toyplot",
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    packages=find_packages(),
    install_requires=[
        "toyplot>=0.18.0",
        "numpy>=1.8.0",
        "future",
        "requests",
        "six",
    ],
    entry_points={},
    license='GPL',
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
)
